#ifndef GUARD_DDT_SLOW_MONOPOLE_TRIGGER_HH
#define GUARD_DDT_SLOW_MONOPOLE_TRIGGER_HH


#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/TriggerDecision.h"
#include "SlowMonopoleTrigger/Track.h"

#include <art/Framework/Principal/Handle.h>
#include <fhiclcpp/ParameterSet.h>

#include <map>
#include <set>
#include <vector>


namespace novaddt
{
  namespace smt
  {
    class Trigger;
  }
}


class novaddt::smt::Trigger
{
public:
  Trigger() = default;
  Trigger(fhicl::ParameterSet const& p);
  
  bool run_algorithm(art::Handle<novaddt::HitList> const& hits);
  std::vector<novaddt::TriggerDecision> trigger_decisions() const;
  
private:
  novaddt::HitList::const_iterator find
  (novaddt::HitList::const_iterator begin,
   novaddt::HitList::const_iterator end, novaddt::TDC tdc) const;

  novaddt::HitList::const_iterator find
  (novaddt::HitList const& hits, novaddt::TDC tdc) const;

  int find_max_gap
  (std::set<int> const& collection) const;

  void find_view_matched_hits
  (std::map<std::string, novaddt::HitList> & hitmap) const;

  bool gap_exceeds_cut
  (std::set<int> collection, int const& object, int const& max_gap_cut) const;

  bool good_track
  (Track const& track) const;
  
  bool hit_is_between_track_end_points
  (novaddt::DAQHit const& hit, Track const& track) const;

  bool hits_are_view_matched
  (novaddt::DAQHit const& x_hit, novaddt::DAQHit const& y_hit) const;

  bool hit_is_in_time_with_road
  (novaddt::DAQHit const& hit, Track const& track) const;

  bool hit_is_on_road
  (novaddt::DAQHit const& hit, Track const& track) const;

  bool hit_is_on_surface
  (novaddt::DAQHit const& hit) const;

  void split_by_view
  (novaddt::HitList hits,
   std::map<std::string, novaddt::HitList> & hitmap) const;

  int max_cell_gap_cut_;
  int max_plane_gap_cut_;
  unsigned sparsification_factor_;
  double track_max_beta_;
  double track_min_beta_;
  unsigned trigger_time_buffer_;
  double trigger_timeout_;
  bool trigger_use_timeout_;
  
  std::vector<novaddt::TriggerDecision> trigger_decisions_;
};


#endif
