////////////////////////////////////////////////////////////////////////
// Class:       MilliBlockChecking
// Module Type: analyzer
// File:        MilliBlockChecking_module.cc
//
// Author: Zukai
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/GroupedHitList.h"

#include <algorithm>
#include <fstream>

namespace novaddt {
  class MilliBlockChecking;
  struct Hit;
  struct CompareHit;
}

class novaddt::MilliBlockChecking : public art::EDAnalyzer {
public:
  explicit MilliBlockChecking(fhicl::ParameterSet const & p);
  virtual ~MilliBlockChecking();

  void analyze(art::Event const & e) override;


private:
  std::string _hitslabel;
  std::string _hitsinstance;
};



novaddt::MilliBlockChecking::MilliBlockChecking(fhicl::ParameterSet const & p) :
  EDAnalyzer(p),
  _hitslabel(p.get< std::string >("hits_label")),
  _hitsinstance(p.get< std::string >("hits_instance"))
{
}

novaddt::MilliBlockChecking::~MilliBlockChecking()
{
}

void novaddt::MilliBlockChecking::analyze(art::Event const & e)
{

  art::Handle<novaddt::HitList> hits;
  e.getByLabel(_hitslabel, _hitsinstance, hits);


  if (hits->size() > 0) {
    uint64_t Tlow  = hits->begin()->TDC().val;
    uint64_t Thigh = (hits->end()-1)->TDC().val;

    std::cout<<Tlow<<"     "<<Thigh<<std::endl;
  }

}

DEFINE_ART_MODULE(novaddt::MilliBlockChecking)
