////////////////////////////////////////////////////////////////////////
// Class:       SNAna
// Module Type: analyzer
// File:        SNAna_module.cc
//
// Generated at Mon Sep  1 04:46:09 2014 by Alec Habig using artmod
// from cetpkgsupport v1_07_00.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// Added after artmod creation
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"
#include "DDTBaseDataProducts/HitList.h"

#include <TH1.h>
#include <TH2.h>
#include <tuple>
// end new includes

namespace novaddt {
  class SNAna;
}

class novaddt::SNAna : public art::EDAnalyzer {
public:
  explicit SNAna(fhicl::ParameterSet const & p);
  // The destructor generated by the compiler is fine for classes
  // without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  SNAna(SNAna const &) = delete;
  SNAna(SNAna &&) = delete;
  SNAna & operator = (SNAna const &) = delete;
  SNAna & operator = (SNAna &&) = delete;

  // Required functions.
  void analyze(art::Event const & e) override;

  // Selected optional functions.
  void beginJob() override;
  void endJob() override;

private:

  // Declare member data here.
  typedef std::tuple<unsigned, double, double> Binning;

  void make_histo(std::string const name, std::string const title, 
		  Binning const& bin);

  std::string slice_label_, slice_instance_;
  int             fMaxPlaneSep;
  int             fMaxTDCSep;
  uint            fMinSliceSize;
  uint            fMaxSliceSize;

  std::map<std::string, TH1*> h_;
  std::map<std::string, TH2*> h2_;
  art::ServiceHandle<art::TFileService> tfs_;

};


novaddt::SNAna::SNAna(fhicl::ParameterSet const & p)
  :
  EDAnalyzer(p),  
 // More initializers here.
  slice_label_    (p.get<std::string        >("slice_label")),
  slice_instance_ (p.get<std::string        >("slice_instance")),
  fMaxPlaneSep    (p.get< int               >("MaxPlaneSep"  )),
  fMaxTDCSep      (p.get< int               >("MaxTDCSep"    )),
  fMinSliceSize   (p.get< int               >("MinSliceSize" )),
  fMaxSliceSize   (p.get< int               >("MaxSliceSize" ))  
{
    std::cout << "--- novaddt::SNAna begin" << std::endl;
    std::cout << "\t Input slices label:      " << slice_label_  << std::endl;
    std::cout << "\t Input slices instance:   " << slice_instance_<< std::endl;
    std::cout << "\t Max Z separation:        " << fMaxPlaneSep   << std::endl;
    std::cout << "\t Max time separation:     " << fMaxTDCSep     << std::endl;
    std::cout << "\t Min slice size:          " << fMinSliceSize  << std::endl;
    std::cout << "\t Max slice size:          " << fMaxSliceSize  << std::endl;
}

void novaddt::SNAna::analyze(art::Event const & e)
{
  // THis Ana module digests slices in the form of vectors of hitlists
  // probably generated by HitSlicers/Multiplet
  art::Handle<std::vector<novaddt::HitList> > slices;
  e.getByLabel(slice_label_, slice_instance_, slices);

  // record number of events (we're simulating 200)
  h_["counter"]->Fill(0);
  h_["counter_precut"]->Fill(0);

  for (auto const& slice : *slices)
  {
    // skip noise slice, the last one
    if (slice == slices->back()) continue;
    // record initial number of slices
    h_["counter_precut"]->Fill(1);
    //    int llasttdc = 0;
    //    int llastplane = 0;
    //    for (auto const& hit : slice) {
      //      if (slice.size()>=40) {
      //	if (llasttdc!=0) {
      //	  int tdcdelta = abs((int)hit.TDC().val-llasttdc);
      //	  std::cout << "next hit: tdc delta:"<< tdcdelta;
      //	  int planedelta = abs((int)hit.Plane().val-llastplane);
      //	  std::cout << " plane delta:" << planedelta<< std::endl;
      //	}
      //	// record the last values of things we're differencing
      //	llasttdc = (int)hit.TDC().val;
      //	llastplane = (int)hit.Plane().val;
      //    }
      //    }    
    double adc_sum = 0.0;

    int lasttdc = 0;
    int lastplane = 0;
    unsigned int slicehits = 0;
    unsigned int slicehits_precut = 0;

    // loop over hits in slice, decide to keep it or not
    bool keepit = true;
    for (auto const& hit : slice) {
      // record all-hits stats
      auto adc = hit.ADC().val;
      h_["adc_all_hits"]->Fill(adc);
      adc_sum += adc;
      slicehits_precut++;
      h_["counter_precut"]->Fill(2);
      h_["plane_all_hits"]->Fill(hit.Plane().val);
      h_["cell_all_hits"]->Fill(hit.Cell().val);

      // if this isn't the first hit in a slice, do stuff
      if (lasttdc !=0 ) {
	int tdcdelta = abs((int)hit.TDC().val-lasttdc);
	//	std::cout << "next hit: tdc delta:"<< tdcdelta;
	int planedelta = abs((int)hit.Plane().val-lastplane);
	//	std::cout << " plane delta:" << planedelta<< std::endl;
	h_["plane_delta_precut"]->Fill((double)planedelta);
	h_["tdc_delta_precut"]->Fill((double)tdcdelta);
	if ((planedelta <= fMaxPlaneSep) && (tdcdelta <= fMaxTDCSep)) 
	  slicehits++;
	else
	  keepit = false;;   // to big a gap in this slice, bail
      }// else std::cout << "first hit"<<std::endl;
      
      // record the last values of things we're differencing
      lasttdc = (int)hit.TDC().val;
      lastplane = (int)hit.Plane().val;
    }
    // plus-1 because we only started counting on hit 2, to get deltas
    if (keepit 
	&& (((slicehits+1)>=fMinSliceSize) && ((slicehits+1)<=fMaxSliceSize))) 
      h_["counter"]->Fill(1);  // record we're keeping a slice
    else 
      keepit = false;
  
    // now go do some stats on the slice if we're keeping it
    slicehits = 0;
    lasttdc = 0;
    lastplane = 0;
    if (keepit) {
      for (auto const& hit : slice) {
	slicehits++;
	h_["counter"]->Fill(2);

	// if this isn't the first hit in a slice, do stuff
	if (lasttdc !=0 ) {
	  int tdcdelta = abs((int)hit.TDC().val-lasttdc);
	  //	std::cout << "next hit: tdc delta:"<< tdcdelta;
	  int planedelta = abs((int)hit.Plane().val-lastplane);
	  //	std::cout << " plane delta:" << planedelta<< std::endl;
	  h_["plane_delta"]->Fill((double)planedelta);
	  h_["tdc_delta"]->Fill((double)tdcdelta);
	}// else std::cout << "first hit"<<std::endl;
      
	// record the last values of things we're differencing
	lasttdc = (int)hit.TDC().val;
	lastplane = (int)hit.Plane().val;
      }
      h_["n_hits_per_slice"]->Fill(double(slicehits));
    }

    double adc_mean;

    if (slice.size() > 0)
    {
      adc_mean = adc_sum / double(slice.size());
    } else {
      adc_mean = 0.0;
    }

    h_["adc_mean"]->Fill(adc_mean);
    h_["adc_sum"]->Fill(adc_sum);

    h_["n_hits_per_slice_precut"]->Fill(double(slicehits_precut));
  }
}

void novaddt::SNAna::beginJob()
{
  // Implementation of optional member function here.
  Binning adc_bin(4096, 0.0, 4096.0);
  Binning nbin(101, 0.0, 100.0);
  Binning plane_bin(1000, 0.0, 1000.0);
  Binning cell_bin(400, -0.0, 400.0);

  make_histo("counter_precut", "Counter (0 = Event, 1 = Slice, 2 = Hit)", 
	     Binning (3,0.0,3.0));
  make_histo("counter", "Counter (0 = Event, 1 = Slice, 2 = Hit)", 
	     Binning (3,0.0,3.0));

  make_histo("adc_all_hits", "ADC Distribution of All Hits", adc_bin);
  make_histo("adc_mean", "Mean ADC per Slice", adc_bin);
  make_histo("adc_sum", "Summed ADC per Slice", adc_bin);

  make_histo("n_hits_per_slice_precut", "Number of Hits per Slice", nbin);
  make_histo("n_hits_per_slice", "Number of Hits per Slice", nbin);

  make_histo("plane_all_hits", "Plane of All Hits", plane_bin);
  make_histo("cell_all_hits", "Cell of All Hits", cell_bin);

  make_histo("tdc_delta_precut","TDC Difference Between Hits", nbin);
  make_histo("plane_delta_precut","Plane Difference Between Hits", nbin);
  make_histo("tdc_delta","TDC Difference Between Hits", nbin);
  make_histo("plane_delta","Plane Difference Between Hits", nbin);
}

void novaddt::SNAna::make_histo
(std::string const name, std::string const title, Binning const& bin)
{
  h_[name] = tfs_->make<TH1F>
    (name.c_str(), title.c_str(), 
     std::get<0>(bin), std::get<1>(bin), std::get<2>(bin));
}

void novaddt::SNAna::endJob()
{
  // Implementation of optional member function here.

  // hmm - we want to write out those histos, yes?

  // blurt out stats
  std::cout << h_["counter_precut"]->GetBinContent(1) << " Pre-cut Events "
	    << h_["counter_precut"]->GetBinContent(2) << " Slices "
	    << h_["counter_precut"]->GetBinContent(3) << " Hits"
	    << std::endl;
  std::cout << h_["counter"]->GetBinContent(1) << " Events "
	    << h_["counter"]->GetBinContent(2) << " Slices "
	    << h_["counter"]->GetBinContent(3) << " Hits"
	    << std::endl;
}

DEFINE_ART_MODULE(novaddt::SNAna)
