////////////////////////////////////////////////////////////////////////
// Class:       RestrictHitsByPlane
// Module Type: filter
// File:        RestrictHitsByPlane_module.cc
//
// Generated at Wed June 5 by Matthew Tamsett based on SpaceSlice
// 
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"

#include "DAQChannelMap/DAQChannelMap.h"
//---------------------------------------------------------------------
namespace novaddt {
    class RestrictHitsByPlane;
}
//---------------------------------------------------------------------
class novaddt::RestrictHitsByPlane : public art::EDFilter{
    public:
        explicit RestrictHitsByPlane(fhicl::ParameterSet const & p);
        virtual ~RestrictHitsByPlane();
        virtual bool filter(art::Event & event);
        void endJob() override;
    private:
        // Declare member data here.
        std::string         fInputLabel;
        std::string         fInputInstance;
        uint                fMaxPlane;
        // Counters
        int _nEvents        = 0;
        int _nInputHits     = 0;
        int _nOutputHits    = 0;
};
//---------------------------------------------------------------------
novaddt::RestrictHitsByPlane::RestrictHitsByPlane(fhicl::ParameterSet const & p)
:
    fInputLabel     (p.get< std::string >("InputLabel"   )),
    fInputInstance  (p.get< std::string >("InputInstance")),
    fMaxPlane       (p.get< uint >       ("MaxPlane"))
    
    // Initialize member data here.
{
    std::cout << "=== novaddt::RestrictHitsByPlane instantiate" << std::endl;
    std::cout << "\t input hits label:          " << fInputLabel      << std::endl;
    std::cout << "\t input hits instance:       " << fInputInstance   << std::endl;
    std::cout << "\t max plane:                 " << fMaxPlane        << std::endl;
    // Call appropriate Produces<>() functions here.
    produces<HitList>();
}
//---------------------------------------------------------------------
novaddt::RestrictHitsByPlane::~RestrictHitsByPlane()
{
    // Clean up dynamic memory and other resources here.
}
//---------------------------------------------------------------------
bool  novaddt::RestrictHitsByPlane::filter(art::Event & event)
{
    //LOG_DEBUG("RestrictHitsByPlane") << "=== novaddt::RestrictHitsByPlane filter. Event: " 
              //<< event.id().event() 
              //<< std::endl;
    _nEvents++;
    art::Handle<HitList> hit_list;
    event.getByLabel(fInputLabel, fInputInstance, hit_list);
    //LOG_DEBUG("RestrictHitsByPlane") << "\t got " << hit_list->size() << " hits" << std::endl;
    
    //make a hit lists out of hits/hit lists that pass the spatial noise removal
    std::unique_ptr< HitList >final_product(new HitList);
    
    for(unsigned int i=0; i<hit_list->size(); ++i){
        _nInputHits++;
        DAQHit product = hit_list->at(i);
        if(product.Plane().val > fMaxPlane){
            continue;
        }
        final_product->push_back(product);
    } // end of loop on hits lists
    //put the hits into the event for later use
    //LOG_DEBUG("RestrictHitsByPlane") << "\t produced " << final_product->size() << " hits" << std::endl;
    _nOutputHits+=final_product->size();
    event.put(std::move(final_product));

    return true;
}
//---------------------------------------------------------------------
void novaddt::RestrictHitsByPlane::endJob()
{
    std::cout << "=== novaddt::RestrictHitsByPlane endJob" << std::endl;
    std::cout << "\tNumber of events:             " << _nEvents       << std::endl;
    std::cout << "\tNumber of hits:               " << _nInputHits    << std::endl;
    std::cout << "\tNumber of hits filtered:      " << _nOutputHits   << std::endl;
}
//---------------------------------------------------------------------
DEFINE_ART_MODULE(novaddt::RestrictHitsByPlane)
