////////////////////////////////////////////////////////////////////////
// Class:       TDSlice
// Module Type: producer
// File:        TDSlice_module.cc
//
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/GroupedHitList.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"

namespace novaddt {
  class TDSlice;
}

class novaddt::TDSlice : public art::EDFilter{
public:
  explicit TDSlice(fhicl::ParameterSet const & p);
  virtual ~TDSlice();

  virtual bool filter(art::Event & e);

private:

  // Declare member data here.
  std::string _slicelabel;
  std::string _sliceinstance;
  //  unsigned int fMinHits;

};


novaddt::TDSlice::TDSlice(fhicl::ParameterSet const & p)
:
  _slicelabel(p.get<std::string>("slice_label")),
  _sliceinstance(p.get<std::string>("slice_instance"))
  // Initialize member data here.
{
  // Call appropriate Produces<>() functions here.
  produces<std::vector<HitList>>("");

}

novaddt::TDSlice::~TDSlice()
{
  // Clean up dynamic memory and other resources here.
}

bool  novaddt::TDSlice::filter(art::Event & e)
{
  art::Handle<std::vector<novaddt::HitList> > slices;
  e.getByLabel(_slicelabel, _sliceinstance, slices);

  //make a grouped hitlist out of hits after time slicing
  std::unique_ptr< std::vector<HitList> >product(new std::vector<HitList>);

  for (unsigned i = 0; i!= slices->size(); ++i) {
    HitList const &hits = slices->at(i);
    HitList xslice, yslice;

    for (auto const hit: hits) {
      if (hit.View().val == daqchannelmap::X_VIEW) {
	xslice.emplace_back(hit);
      }
      else yslice.emplace_back(hit);
    }
    
    product->emplace_back(xslice);
    product->emplace_back(yslice);
    
 }

  //put the hits into the event for later use
  e.put(std::move(product),"");

  return true;
}


DEFINE_ART_MODULE(novaddt::TDSlice)
