////////////////////////////////////////////////////////////////////////
// Class:       HighSortByTDC
// Module Type: producer
// File:        HighSortByTDC_module.cc
//
// Generated at Thu Oct 18 16:45:39 2012 by Martin Frank using artmod
// from art v1_01_01.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"

#include "DDTBaseDataProducts/CompareDAQHit.h"
#include "DDTBaseDataProducts/HitList.h"

namespace novaddt {
  class HighSortByTDC;
}

class novaddt::HighSortByTDC : public art::EDFilter {
public:
  explicit HighSortByTDC(fhicl::ParameterSet const & p);
  virtual ~HighSortByTDC();

  virtual bool filter(art::Event & e);


private:

  std::string _hitslabel;
  int _ADCmin;

};


novaddt::HighSortByTDC::HighSortByTDC(fhicl::ParameterSet const & p)
  : _hitslabel(p.get< std::string >("hits_label")),
    _ADCmin(p.get<int>("ADC_min"))
{
  produces<HitList>("TDCSortedHighHits");
}

novaddt::HighSortByTDC::~HighSortByTDC()
{
  // Clean up dynamic memory and other resources here.
}

bool novaddt::HighSortByTDC::filter(art::Event & e)
{
  art::Handle<novaddt::HitList> hits;
  e.getByLabel(_hitslabel, hits);
  //  HitList hit_list(*hits);

  HitList hit_list;
  for (auto hit: *hits) {
    if (hit.ADC().val>_ADCmin) hit_list.emplace_back(hit);
  }

  if (hit_list.size()==0) return false;
  sort(hit_list.begin(), hit_list.end(), CompareDAQHit<TDC>());

  std::unique_ptr<HitList> product(new HitList(hit_list));
  e.put(std::move(product), "TDCSortedHighHits");

  return true;
}

DEFINE_ART_MODULE(novaddt::HighSortByTDC)
