////////////////////////////////////////////////////////////////////////
// Class:       ContainedSlice
// Module Type: filter
// File:        ContainedSlice_module.cc
//
// Generated at Tue May 21 09:39:04 2013 by Matthew Tamsett using artmod
// from art v1_02_06.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"
#include "DDTBaseDataProducts/GroupedHitList.h"        
#include "DAQChannelMap/DAQChannelMap.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "TVector3.h"

//---------------------------------------------------------------------
namespace novaddt {
    class ContainedSlice;
}
//---------------------------------------------------------------------
class novaddt::ContainedSlice : public art::EDFilter {
    public:
        explicit ContainedSlice(fhicl::ParameterSet const & p);
        virtual ~ContainedSlice();
        bool filter(art::Event & event) override;
        bool beginRun(art::Run & r) override;
        bool beginSubRun(art::SubRun & sr) override;
        void endJob() override;
        void printDDTSlices(art::Handle< std::vector<novaddt::HitList> > & ddt_slices);
        void findBoundingBox(novaddt::HitList & xz_hits, novaddt::HitList & yz_hits, 
                             TVector3 & start, TVector3 & end);
        void filterForContainedHitLists(art::Handle< std::vector<novaddt::HitList> > & ddt_slices, 
                                        std::unique_ptr< std::vector<HitList> > & contained_hitlists);
        bool defineContainmentRegion(art::Event & event);
        void zeroCuts();
    private:
        // FHiCL configuration options
        std::string  _sliceModuleLabel;   ///< label of module making the HitList
        std::string  _sliceInstanceLabel; ///< instance label making the HitList
        float _xMin;                      ///< lower x fiducial cut
        float _xMax;                      ///< upper x fiducial cut
        float _yMin;                      ///< lower y fiducial cut
        float _yMax;                      ///< upper y fiducial cut
        unsigned int _zMin;               ///< lower z fiducial cut
        unsigned int _zMax;               ///< upper z fiducial cut
        unsigned int _secondaryZMin;      ///< optional secondary lower z fiducial cut
        unsigned int _secondaryZMax;      ///< optional secondary upper z fiducial cut
        bool _useSecondaryZCuts;          ///< Apply the optional secondary z cuts
        bool _autoconfigure;              ///< Autoconfigure z containment cuts
        unsigned int _autoconfigureMinZOffset; ///< The distance from the lower edge set by autoconfigure
        unsigned int _autoconfigureMaxZOffset; ///< The distance from the upper edge set by autoconfigure
        unsigned int _autoconfigureInt;   ///< Run autoconfigure after every this number of events
        std::string _detector;            ///< Detector ID, can be one of FD, ND or NDOS
        int _detector_id;                 ///< Detector ID int used to configure DAQ channel map
        std::string _autoconfigureSliceModuleLabel;   ///< label of module making the HitList for use in auto configuration
        std::string _autoconfigureSliceInstanceLabel; ///< instance label making the HitList for use in auto configuration
        bool _verbose;                    ///< Turn on verbose mode
        
        bool _reconfigure = true;         ///< Force a reconfigure if autoconfigure is enabled
        unsigned int _nDCMsPerDiBlock;    ///< Number of DCMs per diblock
        unsigned int _nPlanesPerDiBlock;  ///< Number of Planes per diblock
        
        unsigned int _nEvents         = 0;///< number of events seen
        unsigned int _nSlices         = 0;///< number of slices seen
        unsigned int _nFilteredSlices = 0;///< number of slices filtered
        unsigned int _nAutoconfigures = 0;///< number of times autoconfigure is run
        unsigned int _nDiblocks       = 0;///< The total number of diblocks found
        unsigned int _nGoodDiblocks   = 0;///< The number of diblocks with all DCMs
};
//---------------------------------------------------------------------
novaddt::ContainedSlice::ContainedSlice(fhicl::ParameterSet const & p)
:   _sliceModuleLabel  (p.get<std::string >("SliceModuleLabel"      ))
  , _sliceInstanceLabel(p.get<std::string >("SliceInstanceLabel", ""))
  , _xMin (p.get<float>("Xmin"))
  , _xMax (p.get<float>("Xmax"))
  , _yMin (p.get<float>("Ymin"))
  , _yMax (p.get<float>("Ymax"))
  , _zMin (p.get<unsigned int>("Zmin"))
  , _zMax (p.get<unsigned int>("Zmax"))
  , _secondaryZMin    (p.get<unsigned int>("SecondaryZmin"))
  , _secondaryZMax    (p.get<unsigned int>("SecondaryZmax"))
  , _useSecondaryZCuts(p.get<bool>("UseSecondaryZCuts"))
  , _autoconfigure    (p.get<bool>("Autoconfigure"))
  , _autoconfigureMinZOffset (p.get<unsigned int>("AutoconfigureMinZOffset"))
  , _autoconfigureMaxZOffset (p.get<unsigned int>("AutoconfigureMaxZOffset"))
  , _autoconfigureInt (p.get<unsigned int>("AutoconfigureInterval"))
  , _detector         (p.get<std::string>("Detector"))
  , _autoconfigureSliceModuleLabel  (p.get<std::string >("AutoconfigureSliceModuleLabel"      ))
  , _autoconfigureSliceInstanceLabel(p.get<std::string >("AutoconfigureSliceInstanceLabel", ""))
  , _verbose(p.get<bool>("Verbose"))
{
    std::cout << "--- novaddt::ContainedSlice instantiate" << std::endl;
    std::cout << "\t SliceModuleLabel:          " << _sliceModuleLabel << std::endl;
    std::cout << "\t SliceInstanceLabel:        " << _sliceInstanceLabel << std::endl;
    std::cout << "\t X fiducial cuts:           " << _xMin << " to " << _xMax << std::endl;
    std::cout << "\t Y fiducial cuts:           " << _yMin << " to " << _yMax << std::endl;
    std::cout << "\t Z fiducial cuts:           " << _zMin << " to " << _zMax << std::endl;
    std::cout << "\t Use secondary z cuts:      " << _useSecondaryZCuts << std::endl;
    std::cout << "\t Secondary Z fiducial cuts: " << _secondaryZMin << " to " << _secondaryZMax << std::endl;
    std::cout << "\t Autoconfigure:             " << _autoconfigure << std::endl;
    std::cout << "\t Autoconfigure min Z offset:" << _autoconfigureMinZOffset << std::endl;
    std::cout << "\t Autoconfigure max Z offset:" << _autoconfigureMaxZOffset << std::endl;
    std::cout << "\t Autoconfigure interval:    " << _autoconfigureInt << std::endl;
    std::cout << "\t Detector:                  " << _detector << std::endl;
    std::cout << "\t AC SliceModuleLabel:       " << _autoconfigureSliceModuleLabel << std::endl;
    std::cout << "\t AC SliceInstanceLabel:     " << _autoconfigureSliceInstanceLabel << std::endl;
    std::cout << "\t Verbose:                   " << _verbose << std::endl;
    
    // Assert that a valid detector has been setup
    assert( (_detector == "FD") ||
            (_detector == "ND") ||
            (_detector == "NDOS") );
    // Set the detector id int
    if(_detector == "FD"){
        _detector_id = novadaq::cnv::kFARDET;
        // These should really be got from DAQ channel map
        _nDCMsPerDiBlock   = 12;
        _nPlanesPerDiBlock = 64;
    } else if (_detector == "ND") {
        _detector_id = novadaq::cnv::kNEARDET;
        _nDCMsPerDiBlock = 4;
        _nPlanesPerDiBlock = 64;
    } else { 
        _detector_id = novadaq::cnv::kNDOS;
        _nDCMsPerDiBlock = 4;
        _nPlanesPerDiBlock = 64;
    }

    produces<std::vector<HitList>>("ContainedTimeSpaceSlice");
}
//---------------------------------------------------------------------
novaddt::ContainedSlice::~ContainedSlice()
{
    // Clean up dynamic memory and other resources here.
}
//---------------------------------------------------------------------
bool novaddt::ContainedSlice::beginRun(art::Run & r)
{
    if (_verbose) std::cout << "ContainedSlice::beginRun, setting reconfigure to true\n";
    _reconfigure = true;
    return true;
}
//---------------------------------------------------------------------
bool novaddt::ContainedSlice::beginSubRun(art::SubRun & sr)
{
    if (_verbose) std::cout << "ContainedSlice::beginSubRun, setting reconfigure to true\n";
    _reconfigure = true;
    return true;
}
//---------------------------------------------------------------------
bool novaddt::ContainedSlice::filter(art::Event & event)
{
    // Implementation of required member function here.
    //std::cout << "=== novaddt::ContainedSlice filter. Event: " 
    //          << event.id().event() 
    //          << std::endl;
    _nEvents++;
    // make a hit list return
    std::unique_ptr< std::vector<HitList> > contained_hitlists(new std::vector<HitList>);
    
    // autoconfigure
    if (_autoconfigure && 
        ( ((_nEvents%_autoconfigureInt) == 0) || (_reconfigure) )
        ){
        bool containment_defined = defineContainmentRegion(event);
        if (!containment_defined) return false;
    }
    
    // get the online ddt slices
    art::Handle< std::vector<novaddt::HitList> > slices;
    event.getByLabel(_sliceModuleLabel, _sliceInstanceLabel, slices);
    //std::cout << "\tgot " << slices->size() << " slices" << std::endl;
    _nSlices+=slices->size();
    
    
    //printDDTSlices(slices);
    
    // filter for contained slices
    filterForContainedHitLists(slices, contained_hitlists);    
    //std::cout << "\trecording " << contained_hitlists->size() << " contained slices" << std::endl;
    
    //put the hits into the event for later use
    _nFilteredSlices+=contained_hitlists->size();
    event.put(std::move(contained_hitlists),"ContainedTimeSpaceSlice");
    
    //if (contained_hitlists->size() == 0) return false;
    return true;
}
//---------------------------------------------------------------------
void novaddt::ContainedSlice::printDDTSlices(art::Handle< std::vector<novaddt::HitList> > & ddt_slices)
{
    // loop over grouped hit list
    for(unsigned int i = 0; i < ddt_slices->size(); ++i){
        novaddt::HitList hits(ddt_slices->at(i));
        // sort them
        sort(hits.begin(), hits.end(), novaddt::CompareDAQHit<novaddt::TDC>());
        // earliest and latest TDC
        int eTDC = hits[0].TDC().val;
        int lTDC = hits[hits.size()-1].TDC().val;
        int ADC_sum = 0.;
        novaddt::HitList xz_hits;
        novaddt::HitList yz_hits;
        // loop over hits in each slice
        for(unsigned int j = 0; j < hits.size(); ++j){
            novaddt::DAQHit hit(hits[j]);
            ADC_sum = ADC_sum + hit.ADC().val;
            std::string view = "y";
            if( hit.View().val == daqchannelmap::X_VIEW ){
                xz_hits.push_back(hit);
                view = "x";
            } else {
                yz_hits.push_back(hit);
            }
            //LOG_DEBUG("ContainedSlice") << "\t\t\thit[" << j
                      //<< "]: TDC: "   << hit.TDC().val
                      //<< ", ADC: "    << hit.ADC().val
                      //<< ", plane: "  << hit.Plane().val
                      //<< ", "         << view
                      //<< "-cell: "    << hit.Cell().val
                      //<< std::endl;            
        } // end of loop on hits
        TVector3 start(0,0,0);
        TVector3 end(0,0,0);
        findBoundingBox(xz_hits, yz_hits, start, end);
        // print them
        LOG_DEBUG("ContainedSlice") << "\t\tHit list[" << i
                  << "]: "           << hits.size()
                  << " hits, TDC: "  << eTDC
                  << " - "           << lTDC
                  << ", ADC sum: "   << ADC_sum
                  << ", x: "         << start.X()
                  << " - "           << end.X()
                  << ", y: "         << start.Y()
                  << " - "           << end.Y()
                  << ", z: "         << start.Z()
                  << " - "           << end.Z()
                  << std::endl;
        
    } // end of loop on sliced hits
}
//---------------------------------------------------------------------
void novaddt::ContainedSlice::findBoundingBox(novaddt::HitList & xz_hits, novaddt::HitList & yz_hits, 
                                                TVector3 & start, TVector3 & end)

{
    //if (_verbose) std::cout << "finding bounding box, " 
                            //<< xz_hits.size() << " xz hits, "
                            //<< yz_hits.size() << " yz hits"
                            //<< std::endl;
    // find z boundaries
    // default to the mid point
    int xz_z_min = 448;
    int xz_z_max = 448;
    int xz_x_min = 192;
    int xz_x_max = 192;
    if (xz_hits.size() > 0){
        xz_z_min = xz_hits[0].Plane().val;
        xz_z_max = xz_hits[0].Plane().val;
        xz_x_min = xz_hits[0].Cell().val;
        xz_x_max = xz_hits[0].Cell().val;
    }
    for (auto const& hit : xz_hits){
        if (hit.Plane().val < xz_z_min) xz_z_min = hit.Plane().val;
        if (hit.Plane().val > xz_z_max) xz_z_max = hit.Plane().val;
        if (hit.Cell().val  < xz_x_min) xz_x_min = hit.Cell().val;
        if (hit.Cell().val  > xz_x_max) xz_x_max = hit.Cell().val;
    }
    
    int yz_z_min = 448;
    int yz_z_max = 448;
    int yz_y_min = 192;
    int yz_y_max = 192;
    if (yz_hits.size() > 0){
        yz_z_min = yz_hits[0].Plane().val;
        yz_z_max = yz_hits[0].Plane().val;
        yz_y_min = yz_hits[0].Cell().val;
        yz_y_max = yz_hits[0].Cell().val;
    }
    for (auto const& hit : yz_hits){
        if (hit.Plane().val < yz_z_min) yz_z_min = hit.Plane().val;
        if (hit.Plane().val > yz_z_max) yz_z_max = hit.Plane().val;
        if (hit.Cell().val  < yz_y_min) yz_y_min = hit.Cell().val;
        if (hit.Cell().val  > yz_y_max) yz_y_max = hit.Cell().val;
    }
    
    int z_min = std::min(xz_z_min,yz_z_min);
    int z_max = std::max(xz_z_max,yz_z_max);
    
    start.SetXYZ( double (xz_x_min),
                  double (yz_y_min),
                  double (z_min));
    end.SetXYZ(   double (xz_x_max),
                  double (yz_y_max),
                  double (z_max));
    return;
}
//---------------------------------------------------------------------
void novaddt::ContainedSlice::filterForContainedHitLists(art::Handle< std::vector<novaddt::HitList> > & ddt_slices, 
                                                                std::unique_ptr< std::vector<HitList> > & contained_hitlists)
{
    // loop over grouped hit list
    for(unsigned int i = 0; i < ddt_slices->size(); ++i){
        novaddt::HitList hits(ddt_slices->at(i));
        // earliest and latest TDC
        // sort them 
        //sort(hits.begin(), hits.end(), novaddt::CompareDAQHit<novaddt::TDC>());
        //int eTDC = hits[0].TDC().val;
        //int lTDC = hits[hits.size()-1].TDC().val;
        //unsigned int eTDC = 0;
        //unsigned int lTDC = 0;
        //unsigned int ADC_sum = 0;
        novaddt::HitList xz_hits;
        novaddt::HitList yz_hits;
        // loop over hits in each slice
        for(unsigned int j = 0; j < hits.size(); ++j){
            novaddt::DAQHit hit(hits[j]);
            //if (j == 0) {
                //eTDC = hit.TDC().val;
                //lTDC = hit.TDC().val;
            //} else {
                //if (hit.TDC().val  < eTDC) eTDC = hit.TDC().val;
                //if (hit.TDC().val  > lTDC) lTDC = hit.TDC().val;
            //}
            //ADC_sum = ADC_sum + hit.ADC().val;
            //std::string view = "y";
            if( hit.View().val == daqchannelmap::X_VIEW ){
                xz_hits.push_back(hit);
                //view = "x";
            } else {
                yz_hits.push_back(hit);
            }
            //LOG_DEBUG("ContainedSlice") << "\t\t\thit[" << j
                      //<< "]: TDC: "   << hit.TDC().val
                      //<< ", ADC: "    << hit.ADC().val
                      //<< ", plane: "  << hit.Plane().val
                      //<< ", "         << view
                      //<< "-cell: "    << hit.Cell().val
                      //<< std::endl;            
        } // end of loop on hits
        TVector3 start(0,0,0);
        TVector3 end(0,0,0);
        findBoundingBox(xz_hits, yz_hits, start, end);
        //LOG_DEBUG("ContainedSlice") << "\t\tHit list[" << i
                  //<< "]: "           << hits.size()
                  //<< " hits, TDC: "  << eTDC
                  //<< " - "           << lTDC
                  //<< ", ADC sum: "   << ADC_sum
                  //<< ", x: "         << start.X()
                  //<< " - "           << end.X()
                  //<< ", y: "         << start.Y()
                  //<< " - "           << end.Y()
                  //<< ", z: "         << start.Z()
                  //<< " - "           << end.Z()
                  //<< std::endl;
        // containment
        // x
        if ((start.X() < _xMin) || (start.X() > _xMax) ||
            (end.X()   < _xMin) || (end.X()   > _xMax)){
            //LOG_DEBUG("ContainedSlice") << "\t\t - failed due to x-containment" << std::endl;
            continue;
        }
        // y
        if ((start.Y() < _yMin) || (start.Y() > _yMax) ||
            (end.Y()   < _yMin) || (end.Y()   > _yMax)){
            //LOG_DEBUG("ContainedSlice") << "\t\t - failed due to y-containment" << std::endl;
            continue;
        }
        // z
        if (_useSecondaryZCuts){
            if (((start.Z() < _zMin) || (start.Z() > _zMax) ||
                 (end.Z()   < _zMin) || (end.Z()   > _zMax)) &&
                ((start.Z() < _secondaryZMin) || (start.Z() > _secondaryZMax) ||
                 (end.Z()   < _secondaryZMin) || (end.Z()   > _secondaryZMax)))
                {
                //LOG_DEBUG("ContainedSlice") << "\t\t - failed due to z-containment" << std::endl;
                continue;
            }
        } else {
            if ((start.Z() < _zMin) || (start.Z() > _zMax) ||
                (end.Z()   < _zMin) || (end.Z()   > _zMax)){
                //LOG_DEBUG("ContainedSlice") << "\t\t - failed due to z-containment" << std::endl;
                continue;
            }
        }
        
        // if we reach here then we're contained
        //LOG_DEBUG("ContainedSlice") << "\t\tslice " << i << " is contained (" << hits.size() << " hits)" << std::endl;
        contained_hitlists->push_back(hits);
    } // end of loop on hit lists
}
//---------------------------------------------------------------------
bool novaddt::ContainedSlice::defineContainmentRegion(art::Event & event)
{
    if (_verbose) std::cout << "--- Defining containment region\n";
    _nAutoconfigures++;
    _reconfigure = false;
    std::vector<daqchannelmap::diblock_t>    active_diblocks;
    std::map <daqchannelmap::diblock_t, std::vector<novaddt::DCM> > dcms_per_diblock;
    
    // get the hits
    // to get the best estimate of active regions the most hits should be used
    // therefore we get the hits after TDC sort
    if (_verbose) std::cout << "\tgetting single hit list" << std::endl;
    art::Handle<novaddt::HitList> hits;
    event.getByLabel(_autoconfigureSliceModuleLabel, _autoconfigureSliceInstanceLabel, hits);
    if (_verbose) std::cout << "\tgot " << hits->size() << " hits\n";
    
    // loop over hits
    for(unsigned int i = 0; i < hits->size(); ++i){
        daqchannelmap::diblock_t diblock = daqchannelmap::DAQChannelMap::getInstance(_detector_id)->getDiBlock(hits->at(i).Chan());
        //daqchannelmap::diblock_t diblock = daqchannelmap::DAQChannelMap::getInstance()->getDiBlock(hits->at(i).Chan());
        novaddt::DCM dcm = hits->at(i).getDCM();
        /*
        std::string view = "y";
        if( hits->at(i).View().val == daqchannelmap::X_VIEW ) view = "x";
        if (_verbose) std::cout << "    DCM: "   << dcm.val 
                  << ", diblock: " << diblock
                  << ", view: "    << view
                  << std::endl;
        */
        
        // Is this a new diblock?
        if(std::find(active_diblocks.begin(), active_diblocks.end(), diblock)==active_diblocks.end()){
            active_diblocks.push_back(diblock);
            dcms_per_diblock[diblock].push_back(dcm);
            
            //if (_verbose) std::cout << "     - Found new diblock: " << diblock 
            //          << ", DCM: " << dcm.val
            //          << std::endl;
        } else {
            // Is this a new DCM?
            // There are only 12 DCMs per diblock on the FD (and less on the other detectors)
            // so only try the find if we haven't reached this number
            if (dcms_per_diblock[diblock].size() == 12) continue;
            
            if(std::find(dcms_per_diblock[diblock].begin(), dcms_per_diblock[diblock].end(), dcm)==dcms_per_diblock[diblock].end()){
                dcms_per_diblock[diblock].push_back(dcm);
                
                //if (_verbose) std::cout << "     - Found a new DCM: " << dcm.val
                //          << ", diblock: " << diblock
                //          << std::endl;
            }
        } // end of is new diblock or DCM
    } // end of loop on hits
    
    // sort the diblocks
    std::sort(active_diblocks.begin(), active_diblocks.end());
    
    if (_verbose) std::cout << "Found: "   << active_diblocks.size()
              << " diblocks" << std::endl;
    std::vector<daqchannelmap::diblock_t> good_diblocks;
    for ( auto db : active_diblocks ){
        _nDiblocks++;
        if (_verbose) std::cout << "    DB: " << db
                  << ", DCMs: " << dcms_per_diblock[db].size()
                  << std::endl;
        // Only accept diblocks where all DCMs were active
        if (dcms_per_diblock[db].size() == _nDCMsPerDiBlock){
            _nGoodDiblocks++;
            good_diblocks.push_back(db);
        }
        //for (auto dcm : dcms_per_diblock[db]){
        //    if (_verbose) std::cout << "     - DCM: " << dcm.val << std::endl;
        //}
    }
    
    if (_verbose) std::cout << "Of these " << good_diblocks.size()
              << " have the correct number of DCMs " << std::endl;
    
    // If there are no good diblocks then we can't define containment
    if (good_diblocks.size() == 0){
        if (_verbose) std::cout << "No good diblocks so containment cannot be defined" << std::endl;
        zeroCuts();
        return false;
    }
    
    // Use the list of good diblocks to define our containment region
    // We expect that this should be divided into one, or two continuous regions
    // We do this by finding edges
    std::vector<daqchannelmap::diblock_t> diblock_edges;
    // Add the leading edge
    diblock_edges.push_back(good_diblocks[0]);
    
    // Iterate over diblocks looking for edges
    // Ignoring the first & last entries since by definition that has to be an edge
    for (unsigned int i=1; i < good_diblocks.size(); ++i){
        if (good_diblocks[i] != (good_diblocks[i-1]+1)){
            diblock_edges.push_back(good_diblocks[i-1]);
            diblock_edges.push_back(good_diblocks[i]);
        }
    }
    
    // Add the trailing edge
    diblock_edges.push_back(good_diblocks[good_diblocks.size()-1]);
    
    if (_verbose){
        std::cout << "Found " << diblock_edges.size() 
                  << " diblock edges" << std::endl;
        for (auto db : diblock_edges) std::cout << "    " << db << std::endl;
    }
    
    // Check that an appropriate number of edges have been found
    if ((diblock_edges.size() != 2) && (diblock_edges.size() != 4)){
        if (_verbose) std::cout << "More than two continuous regions found. Containment cannot be defined" << std::endl;
        zeroCuts();
        return false;
    }
    
    // Set the containment cuts
    // Recall that the diblock convention starts from 1
    // Lower edges are therefore set from diblock number -1
    // Upper edges from diblock number
    _zMin = ((diblock_edges[0]-1)*_nPlanesPerDiBlock) + _autoconfigureMinZOffset;
    _zMax = (diblock_edges[1]*_nPlanesPerDiBlock) - _autoconfigureMaxZOffset;
    if (diblock_edges.size() == 4){
        _useSecondaryZCuts = true;
        _secondaryZMin = ((diblock_edges[2]-1)*_nPlanesPerDiBlock) + _autoconfigureMinZOffset;
        _secondaryZMax = (diblock_edges[3]*_nPlanesPerDiBlock) - _autoconfigureMaxZOffset;
    } else {
        _useSecondaryZCuts = false;
    }
    
    if (_verbose){
        std::cout << "Autoconfigured fiducial volume to: " << std::endl;
        std::cout << "\t Z fiducial cuts:           " << _zMin << " to " << _zMax << std::endl;
        std::cout << "\t Use secondary z cuts:      " << _useSecondaryZCuts << std::endl;
        std::cout << "\t Secondary Z fiducial cuts: " << _secondaryZMin << " to " << _secondaryZMax << std::endl;
    }
    
    return true;
}
//---------------------------------------------------------------------
void novaddt::ContainedSlice::zeroCuts()
{
    _zMin = 0;
    _zMax = 0;
    _useSecondaryZCuts = false;
    _secondaryZMin = 0;
    _secondaryZMax = 0;
    _reconfigure   = true;
    if (_verbose){
        std::cout << "Zero-ed fiducial volume to: " << std::endl;
        std::cout << "\t Z fiducial cuts:           " << _zMin << " to " << _zMax << std::endl;
        std::cout << "\t Use secondary z cuts:      " << _useSecondaryZCuts << std::endl;
        std::cout << "\t Secondary Z fiducial cuts: " << _secondaryZMin << " to " << _secondaryZMax << std::endl;
    }
}
//---------------------------------------------------------------------
void novaddt::ContainedSlice::endJob()
{
    std::cout << "--- novaddt::ContainedSlice endJob" << std::endl;
    std::cout << "\tNumber of events:           " << _nEvents         << std::endl;
    std::cout << "\tNumber of slices:           " << _nSlices         << std::endl;
    std::cout << "\tNumber of slices filtered:  " << _nFilteredSlices << std::endl;
    std::cout << "\tNumber of autoconfigures:   " << _nAutoconfigures << std::endl;
    std::cout << "\tNumber of DBs with any DCMs:" << _nDiblocks       << std::endl;
    std::cout << "\tNumber of DBs with all DCMs:" << _nGoodDiblocks   << std::endl;
}
//---------------------------------------------------------------------
DEFINE_ART_MODULE(novaddt::ContainedSlice)
