////////////////////////////////////////////////////////////////////////
// Class:       NuMuTrigger
// Module Type: filter
// File:        NuMuTrigger_module.cc
//
// Generated at Wed May  8 14:35:56 2013 by Matthew Tamsett using artmod
// from art v1_02_06.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/FindOneP.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Persistency/Common/Assns.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/Track3D.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/TriggerDecision.h"

#include "DAQChannelMap/DAQChannelMap.h"
#include "DDTUtilities/AssociationUtil.h"

#include "TVector3.h"

//---------------------------------------------------------------------
namespace novaddt {
    class NuMuTrigger;
}
//---------------------------------------------------------------------
class novaddt::NuMuTrigger : public art::EDFilter {
    public:
        explicit NuMuTrigger(fhicl::ParameterSet const & p);
        virtual ~NuMuTrigger();
        bool    filter(art::Event & e) override;
        void findBoundingBox(novaddt::HitList & xz_hits, novaddt::HitList & yz_hits, 
                             TVector3 & start, TVector3 & end);
        double  n3DCellsToEdge(novaddt::Track3D & track, bool entering);
        void    endJob() override;
    private:
        unsigned int _prescale;           ///< prescale factor (1 out of every this many passes are issued as real triggers)
        std::string  _trackModuleLabel;   ///< label of module making the Tracks
        std::string  _trackInstanceLabel; ///< instance label making the Tracks
        std::string  _sliceModuleLabel;   ///< label of module making the HitList
        std::string  _sliceInstanceLabel; ///< instance label making the HitList
        std::string  _trackToSliceInstanceLabel;///< instance of track to slice association produced by track module
        unsigned int _minPlanesCrossed;   ///< minimum planes crossed for a track to be used
        unsigned int _minCellsPerView;    ///< minimum number of cells per view for a good track
        double       _minCellsToEdge;     ///< containment cut based on minimum number of cell projected to the edge of the detector
        double       _sliceBoxXMin;       ///< lower x slice box fiducial cut
        double       _sliceBoxXMax;       ///< upper x slice box fiducial cut
        double       _sliceBoxYMin;       ///< lower y slice box fiducial cut
        double       _sliceBoxYMax;       ///< upper y slice box fiducial cut
        double       _sliceBoxZMin;       ///< lower z slice box fiducial cut
        double       _sliceBoxZMax;       ///< upper z slice box fiducial cut
        double       _trackBoxXMin;       ///< lower x track box fiducial cut
        double       _trackBoxXMax;       ///< upper x track box fiducial cut
        double       _trackBoxYMin;       ///< lower y track box fiducial cut
        double       _trackBoxYMax;       ///< upper y track box fiducial cut
        double       _trackBoxZMin;       ///< lower z track box fiducial cut
        double       _trackBoxZMax;       ///< upper z track box fiducial cut
        double       _angularCut;         ///< cosmic rejection cut on angle
        double       _enteringCut;        ///< minimum plane crossed to be defined as entering
        double       _cosmicPIDCut;       ///< cosmic pid (MinCellsToEdge/100)^2 + cos(beam)^2 - (dY_unit*4)^2
        double       _angleVarCut;        ///< cosmic rejection cut on angle: cos(beam)^2 - (dY_unit)^2
        bool         _checkProjectionContainment;
        bool         _checkLowZContainment;
        bool         _checkBoxContainment;
        bool         _checkSliceContainment;
        bool         _checkAngle;
        bool         _checkEntering;
        bool         _checkPID;
        bool         _checkAngleVar;
        std::string  _detector;           ///< Set detector specific configuration options, such as the location of the detector boundaries
                                          ///< available options: NDOS, FD, FD_2DB - ND will be added later
        bool         _verbose;
        unsigned int _nEvents           = 0;
        unsigned int _nSlices           = 0;
        unsigned int _nTracks           = 0;
        unsigned int _n3DTracks         = 0;
        unsigned int _nPass_slice_containment = 0;
        unsigned int _nPass_3DTracks    = 0;
        unsigned int _nPass_dX          = 0;
        unsigned int _nPass_dY          = 0;
        unsigned int _nPass_dZ          = 0;
        unsigned int _nPass_projection_containment = 0;
        unsigned int _nPass_box_containment        = 0;
        unsigned int _nPass_angle       = 0;
        unsigned int _nPass_entering    = 0;
        unsigned int _nPass_cosmic_PID  = 0;
        unsigned int _nPass_angle_var   = 0;
        unsigned int _trigger_counts    = 0;
        unsigned int _prescaled_trigger_counts = 0;
};
//---------------------------------------------------------------------
novaddt::NuMuTrigger::NuMuTrigger(fhicl::ParameterSet const & p)
:   _prescale          (p.get<unsigned int>("prescale"              ))
  , _trackModuleLabel  (p.get<std::string >("TrackModuleLabel"      ))
  , _trackInstanceLabel(p.get<std::string >("TrackInstanceLabel", ""))
  , _sliceModuleLabel  (p.get<std::string >("SliceModuleLabel"      ))
  , _sliceInstanceLabel(p.get<std::string >("SliceInstanceLabel", ""))
  , _trackToSliceInstanceLabel(p.get<std::string >("TrackToSliceInstanceLabel", ""))
  , _minPlanesCrossed  (p.get<unsigned int>("MinPlanesCrossed"      ))
  , _minCellsPerView   (p.get<unsigned int>("MinCellsPerView"       ))
  , _minCellsToEdge    (p.get<double>      ("MinCellsToEdge"        ))
  , _sliceBoxXMin      (p.get<double>      ("SliceBoxXMin"          ))
  , _sliceBoxXMax      (p.get<double>      ("SliceBoxXMax"          ))
  , _sliceBoxYMin      (p.get<double>      ("SliceBoxYMin"          ))
  , _sliceBoxYMax      (p.get<double>      ("SliceBoxYMax"          ))
  , _sliceBoxZMin      (p.get<double>      ("SliceBoxZMin"          ))
  , _sliceBoxZMax      (p.get<double>      ("SliceBoxZMax"          ))
  , _trackBoxXMin      (p.get<double>      ("TrackBoxXMin"          ))
  , _trackBoxXMax      (p.get<double>      ("TrackBoxXMax"          ))
  , _trackBoxYMin      (p.get<double>      ("TrackBoxYMin"          ))
  , _trackBoxYMax      (p.get<double>      ("TrackBoxYMax"          ))
  , _trackBoxZMin      (p.get<double>      ("TrackBoxZMin"          ))
  , _trackBoxZMax      (p.get<double>      ("TrackBoxZMax"          ))
  , _angularCut        (p.get<double>      ("AngularCut"            ))
  , _enteringCut       (p.get<double>      ("EnteringCut"           ))
  , _cosmicPIDCut      (p.get<double>      ("CosmicPID"             ))
  , _angleVarCut       (p.get<double>      ("AngleVarCut"           ))
  , _checkProjectionContainment  (p.get<bool>   ("CheckProjectionContainment"  ))
  , _checkLowZContainment        (p.get<bool>   ("CheckLowZContainment"        ))
  , _checkBoxContainment         (p.get<bool>   ("CheckBoxContainment"         ))
  , _checkSliceContainment       (p.get<bool>   ("CheckSliceContainment"       ))
  , _checkAngle        (p.get<bool>        ("CheckAngle"            ))
  , _checkEntering     (p.get<bool>        ("CheckEntering"         ))
  , _checkPID          (p.get<bool>        ("CheckPID"              ))
  , _checkAngleVar     (p.get<bool>        ("CheckAngleVar"         ))
  , _detector          (p.get<std::string> ("Detector"              ))
  , _verbose           (p.get<bool>        ("Verbose"               ))
{
    std::cout << "=== novaddt::NuMuTrigger instantiate" << std::endl;
    std::cout << "\t prescale:                      " << _prescale              << std::endl;
    std::cout << "\t TrackModuleLabel:              " << _trackModuleLabel      << std::endl;
    std::cout << "\t TrackInstanceLabel:            " << _trackInstanceLabel    << std::endl;
    std::cout << "\t SliceModuleLabel:              " << _sliceModuleLabel      << std::endl;
    std::cout << "\t SliceInstanceLabel:            " << _sliceInstanceLabel    << std::endl;
    std::cout << "\t TrackToSliceInstanceLabel:     " << _trackToSliceInstanceLabel << std::endl;
    std::cout << "\t Cut on slice containment:      " << _checkSliceContainment << std::endl;
    std::cout << "\t     - X slice cuts:            " << _sliceBoxXMin << " to " << _sliceBoxXMax << std::endl;
    std::cout << "\t     - Y slice cuts:            " << _sliceBoxYMin << " to " << _sliceBoxYMax << std::endl;
    std::cout << "\t     - Z slice cuts:            " << _sliceBoxZMin << " to " << _sliceBoxZMax << std::endl;
    std::cout << "\t MinCellsPerView:               " << _minCellsPerView       << std::endl;
    std::cout << "\t MinPlanesCrossed:              " << _minPlanesCrossed      << std::endl;
    std::cout << "\t Cut on projection containment: " << _checkProjectionContainment    << std::endl;
    std::cout << "\t     - Detector:                " << _detector              << std::endl;
    std::cout << "\t     - MinCellsToEdge:          " << _minCellsToEdge        << std::endl;
    std::cout << "\t     - low Z containment:       " << _checkLowZContainment  << std::endl;
    std::cout << "\t Cut on track box containment:  " << _checkBoxContainment   << std::endl;
    std::cout << "\t     - X box cuts:              " << _trackBoxXMin << " to " << _trackBoxXMax << std::endl;
    std::cout << "\t     - Y box cuts:              " << _trackBoxYMin << " to " << _trackBoxYMax << std::endl;
    std::cout << "\t     - Z box cuts:              " << _trackBoxZMin << " to " << _trackBoxZMax << std::endl;
    std::cout << "\t Cut on angle:                  " << _checkAngle            << std::endl;
    std::cout << "\t     - Angular cut:             " << _angularCut            << std::endl;
    std::cout << "\t Cut on entering:               " << _checkEntering         << std::endl;
    std::cout << "\t     - Entering cut:            " << _enteringCut           << std::endl;
    std::cout << "\t Cut on pid:                    " << _checkPID              << std::endl;
    std::cout << "\t     - Cosmic PID cut           " << _cosmicPIDCut          << std::endl;
    std::cout << "\t Cut on angle variable:         " << _checkAngleVar         << std::endl;
    std::cout << "\t     - Angle Var cut:           " << _angleVarCut           << std::endl;
    std::cout << "\t Verbose:                       " << _verbose               << std::endl;
    
    // check that the requested detector is configured
    assert( (_detector == "NDOS") ||
            (_detector == "FD")   ||
            (_detector == "FD_2DB")
          );
          
    produces< std::vector<novaddt::TriggerDecision>>();
    produces< art::Assns<novaddt::TriggerDecision, novaddt::HitList> >();
}
//---------------------------------------------------------------------
novaddt::NuMuTrigger::~NuMuTrigger()
{
    // Clean up dynamic memory and other resources here.
}
//---------------------------------------------------------------------
bool novaddt::NuMuTrigger::filter(art::Event & event)
{
    // Implementation of required member function here.
    if (_verbose) std::cout<< "=== novaddt::NuMuTrigger filter. Event: " 
              << event.id().event() 
              << std::endl;
    // make the trigger decision and assignment vectors we are going to write out to the event
    std::unique_ptr<std::vector<novaddt::TriggerDecision>> trigger_decisions(new std::vector<novaddt::TriggerDecision>);
    std::unique_ptr<art::Assns<novaddt::TriggerDecision, novaddt::HitList> > assn(new art::Assns<novaddt::TriggerDecision, novaddt::HitList >);

    // get the slices
    art::Handle< std::vector<novaddt::HitList> > slices;
    event.getByLabel(_sliceModuleLabel, _sliceInstanceLabel, slices);
    if (_verbose) std::cout<< "\tgot " << slices->size() << " slices" << std::endl;
    _nSlices+=slices->size();
    
    // get the tracks for each hit list
    const std::string the_slices = _trackModuleLabel+":"+_trackToSliceInstanceLabel;
    art::FindManyP<novaddt::Track3D> find_tracks(slices, event, the_slices);
    assert(find_tracks.isValid());
    
    // loop over slices
    for(size_t i_slice = 0; i_slice < slices->size(); ++i_slice){
        // get the hit list
        novaddt::HitList hit_list = slices->at(i_slice);
        if (_verbose){
            std::cout<< "\tslice[" << i_slice << "]: " << hit_list.size() << " hits" << std::endl;
            // loop over hits, useful for debugging only at the moment
            for(size_t i_hit = 0; i_hit < hit_list.size(); ++i_hit){
                novaddt::DAQHit hit(hit_list[i_hit]);
                std::string view = "y";
                if( hit.View().val == daqchannelmap::X_VIEW ) view = "x";
                std::cout<< "\t\thit[" << i_hit
                          << "]: TDC: "   << hit.TDC().val
                          << ", ADC: "    << hit.ADC().val
                          << ", plane: "  << hit.Plane().val
                          << ", "         << view
                          << "-cell: "    << hit.Cell().val
                          << std::endl;         
            } // end of loop on hits
        }
        
        //---- check the trigger conditions
        
        // slice box containment
        if (_checkSliceContainment){
            novaddt::HitList xz_hits;
            novaddt::HitList yz_hits;
            // loop over hits in each slice
            for(unsigned int j = 0; j < hit_list.size(); ++j){
                novaddt::DAQHit hit(hit_list[j]);
                if( hit.View().val == daqchannelmap::X_VIEW ){
                    xz_hits.push_back(hit);
                } else {
                    yz_hits.push_back(hit);
                }
            } // end of loop on hits
            TVector3 slice_start(0,0,0);
            TVector3 slice_end(0,0,0);
            findBoundingBox(xz_hits, yz_hits, slice_start, slice_end);
            // x
            if ((slice_start.X() < _sliceBoxXMin) || (slice_start.X() > _sliceBoxXMax) ||
                (slice_end.X()   < _sliceBoxXMin) || (slice_end.X()   > _sliceBoxXMax)){
                if (_verbose) std::cout << "\t\t - failed due to slice x-containment" << std::endl;
                continue;
            }
            // y
            if ((slice_start.Y() < _sliceBoxYMin) || (slice_start.Y() > _sliceBoxYMax) ||
                (slice_end.Y()   < _sliceBoxYMin) || (slice_end.Y()   > _sliceBoxYMax)){
                if (_verbose) std::cout << "\t\t - failed due to slice y-containment" << std::endl;
                continue;
            }
            // z
            if ((slice_start.Z() < _sliceBoxZMin) || (slice_start.Z() > _sliceBoxZMax) ||
                (slice_end.Z()   < _sliceBoxZMin) || (slice_end.Z()   > _sliceBoxZMax)){
                if (_verbose) std::cout << "\t\t - failed due to slice z-containment" << std::endl;
                continue;
            }
        } // end of check on slice containment
        _nPass_slice_containment++;
        
        // find the tracks for this hit list
        std::vector< art::Ptr<novaddt::Track3D> > tracks = find_tracks.at(i_slice);
        if (_verbose) std::cout<< "\t - got " << tracks.size() << " tracks" << std::endl;
        if (!tracks.size() ){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to lack of tracks" << std::endl;
            continue;
        }
        _nTracks+=tracks.size();
        
        // get the hit list for each track
        art::FindOneP<novaddt::HitList> fohl(tracks, event, _trackModuleLabel);
        assert(fohl.isValid());
        assert(fohl.size() == tracks.size());
        
        double longest_length = -9;
        unsigned int i_longest = 0;
        bool track_found = false;
        
        for(size_t i_track = 0; i_track < tracks.size(); ++i_track){
            art::Ptr<novaddt::Track3D> track = tracks[i_track];
            // get the hit list for this track
            art::Ptr<novaddt::HitList> this_hit_list = fohl.at(i_track);
                        
            TVector3 length = track->End() - track->Start();
            double this_length = std::abs(length.Mag());            
                        
            if (_verbose) std::cout<< "\t\t3D-track[" << i_track
                      << "]\t view: " << track->View()
                      << ", hits: "   << this_hit_list->size() 
                      << ", X: "      << track->Start().X() << " - " << track->End().X()
                      << ", Y: "      << track->Start().Y() << " - " << track->End().Y()
                      << ", Z: "      << track->Start().Z() << " - " << track->End().Z()
                      << ", length: " << this_length
                      << std::endl;
                      
            //// loop over hits, useful for debugging only at the moment
            //if (_verbose){
                //for(size_t i_hit = 0; i_hit < this_hit_list->size(); ++i_hit){
                    //std::string h_view = "y";
                    //if( this_hit_list->at(i_hit).View().val == daqchannelmap::X_VIEW) h_view = "x";
                    //std::cout<< "\t\t\thit["   << i_hit
                              //<< "]: \tplane: " << this_hit_list->at(i_hit).Plane().val
                              //<< ",\tcell: "    << this_hit_list->at(i_hit).Cell().val
                              //<< ",\t view: "   << h_view
                              //<< ",\tTDC: "     << this_hit_list->at(i_hit).TDC().val
                              //<< ",\tADC: "     << this_hit_list->at(i_hit).ADC().val
                              //<< std::endl;
                //} // end of loop on hits
            //}
            
            // only select 3D tracks
            if (!track->Is3D()) continue;
            _n3DTracks++;
            // select longest track
            if (this_length > longest_length){
                longest_length = this_length;
                i_longest = i_track;
                track_found = true;
            }
        } // end of loop on tracks
        
        if (!track_found){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due lack of 3D tracks" << std::endl;
            continue;
        }
        _nPass_3DTracks++;
        novaddt::Track3D track = *tracks[i_longest];
        // get the hit list for this track
        art::Ptr<novaddt::HitList> hits_on_track = fohl.at(i_longest);
        
        if (_verbose) std::cout<< "\t\tlongest track[" << i_longest
                  << "], hits: "      << hits_on_track->size() 
                  << ", x: "          << track.Start().X()    << " - " << track.End().X()
                  << ", y: "          << track.Start().Y()    << " - " << track.End().Y()
                  << ", z: "          << track.Start().Z()    << " - " << track.End().Z()
                  << std::endl;
                  
        // check the trigger conditions
        // track exists in both views
        unsigned int x_hits = 0;
        unsigned int y_hits = 0;
        for(size_t i_hit = 0; i_hit < hits_on_track->size(); ++i_hit){
            if( hits_on_track->at(i_hit).View().val == daqchannelmap::X_VIEW){
                x_hits++;
            } else {
                y_hits++;
            } 
        } // end of loop on hits
        // dX
        if (x_hits < _minCellsPerView){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due lack of x hits" << std::endl;
            continue;
        }
        _nPass_dX++;
        // dY
        if (y_hits < _minCellsPerView){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due lack of y hits" << std::endl;
            continue;
        }
        _nPass_dY++;
        // dZ
        double dZ = std::fabs(track.End().Z() - track.Start().Z());
        if (dZ < _minPlanesCrossed){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to dZ" << std::endl;
            continue;
        }
        _nPass_dZ++;
        // projection containment
        double minCellsToEdge = n3DCellsToEdge(track, ((!_checkLowZContainment)||_checkEntering));
        if (_verbose) std::cout<< "\t\t - min distance to any edge: " << minCellsToEdge << std::endl;
        if (_checkProjectionContainment && (minCellsToEdge < _minCellsToEdge)){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to containment cut" << std::endl;
            continue;
        }
        _nPass_projection_containment++;
        // box containment
        if (_checkBoxContainment && (std::min(track.Start().X(), track.End().X()) < _trackBoxXMin)){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to X min containment cut" << std::endl;
            continue;
        }
        if (_checkBoxContainment && (std::min(track.Start().Y(), track.End().Y()) < _trackBoxYMin)){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to Y min containment cut" << std::endl;
            continue;
        }
        if (_checkBoxContainment && (std::min(track.Start().Z(), track.End().Z()) < _trackBoxZMin)){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to Z min containment cut" << std::endl;
            continue;
        }
        if (_checkBoxContainment && (std::max(track.Start().X(), track.End().X()) > _trackBoxXMax)){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to X max containment cut" << std::endl;
            continue;
        }
        if (_checkBoxContainment && (std::max(track.Start().Y(), track.End().Y()) > _trackBoxYMax)){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to Y max containment cut" << std::endl;
            continue;
        }
        if (_checkBoxContainment && (std::max(track.Start().Z(), track.End().Z()) > _trackBoxZMax)){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to Z max containment cut" << std::endl;
            continue;
        }
        _nPass_box_containment++;
        // angle
        double scaled_dZ = (dZ * 6.7) / 4.; // cells are longer in z than in x or y
        double dX = std::fabs( track.End().X() - track.Start().X() );
        double dY = std::fabs( track.End().Y() - track.Start().Y() );
        double CosBeam = ( scaled_dZ/ std::sqrt(dX*dX+dY*dY+scaled_dZ*scaled_dZ) );
        if (_verbose) std::cout<< "\t\t - cos(beam): " << CosBeam << std::endl;
        if (_checkAngle && (CosBeam < _angularCut)){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to CosBeam cut" << std::endl;
            continue;
        }
        _nPass_angle++;
        // entering
        double min_Z = std::min(track.Start().Z(), track.End().Z());
        if (_verbose) std::cout<< "\t\t - min Z: " << min_Z << std::endl;
        if(_checkEntering && (min_Z > _enteringCut)){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to Entering cut" << std::endl;
            continue;
        }
        _nPass_entering++;
        // cosmic PID
        TVector3 track_3d = TVector3();
        track_3d.SetXYZ(dX,dY,scaled_dZ);
        double dY_unit = track_3d.Unit().Y();
        double cosmic_pid = ((minCellsToEdge/100.)*(minCellsToEdge/100.)) + 
                            (CosBeam*CosBeam) - 
                            (dY_unit*dY_unit);
        if (_verbose) std::cout<< "\t\t - cosmic PID: " << cosmic_pid 
                  << ", from nCells: "    << minCellsToEdge
                  << ", cos(beam): "      << CosBeam
                  << ", dY unit: "        << dY_unit
                  << std::endl;
        if (_checkPID && (cosmic_pid < _cosmicPIDCut)){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to cosmic PID cut" << std::endl;
            continue;
        }
        _nPass_cosmic_PID++;
        // angle variable
        // 2-D cut in angle var vs n hits space 
        //double hit_ratio = double(hits_on_track->size()) / double(hit_list.size());
        //if (_checkAngleVar && (hit_ratio > 0.88)){
            //double angle_var = (CosBeam*CosBeam) - (dY_unit*dY_unit);
            //if( hit_list.size() >50  && 
                //hit_list.size()<=135 &&
                //(angle_var<((1.1/85.)*(hit_list.size()-1.65)))
                //){
                //if (_verbose) std::cout<< "\t\t - failed the trigger decision due to angle variable cut" << std::endl;
                //continue;
            //}
            //if( hit_list.size() >135  && 
                //hit_list.size()<=250 &&
                //(angle_var<((0.6/115.)*(hit_list.size()-0.6)))
                //){
                //if (_verbose) std::cout<< "\t\t - failed the trigger decision due to angle variable cut" << std::endl;
                //continue;
            //}
            //if( hit_list.size() >250  && 
                //hit_list.size()<=500 &&
                //(angle_var<((0.1/250.)*(hit_list.size()+0.6)))
                //){
                //if (_verbose) std::cout<< "\t\t - failed the trigger decision due to angle variable cut" << std::endl;
                //continue;
            //}
            //if( hit_list.size() >500 &&
                //(angle_var<0.8)
                //){
                //if (_verbose) std::cout<< "\t\t - failed the trigger decision due to angle variable cut" << std::endl;
                //continue;
            //}
        //}
        // 1-D cut on angle var
        double angle_var = (CosBeam*CosBeam) - (dY_unit*dY_unit);
        if (_checkAngleVar && (angle_var < _angleVarCut)){
            if (_verbose) std::cout<< "\t\t - failed the trigger decision due to angle variable cut" << std::endl;
            continue;
        }
        _nPass_angle_var++;
        // if we reach here then we've passed the trigger
        if (_verbose) std::cout<< "\t\t - passed the trigger" << std::endl;
        _trigger_counts++;
        // check the prescale
        if (_trigger_counts%_prescale == _prescale-1) {
            _prescaled_trigger_counts++;
            if (_verbose) std::cout<< "\t\t - not prescaled" << std::endl;
            // get the min and max TDC from the parent slice 
            // and use these to define the trigger window
            // safest not to assume this is time sorted
            uint64_t min_tdc = hit_list.front().TDC().val;
            uint64_t max_tdc = hit_list.back().TDC().val;
            for(size_t i_hit = 0; i_hit < hit_list.size(); ++i_hit){
                if (hit_list[i_hit].TDC().val < min_tdc) min_tdc = hit_list[i_hit].TDC().val;
                if (hit_list[i_hit].TDC().val > max_tdc) max_tdc = hit_list[i_hit].TDC().val;
            } // end of loop on hits
            
            trigger_decisions->push_back(novaddt::TriggerDecision(
                                         min_tdc, 
                                         max_tdc - min_tdc, 
                                         novaddt::TriggerMask(daqdataformats::trigID::TRIG_ID_DATA_NUMU),
                                         _prescale)
                                         );
            // get the Ptr for the slice
            art::Ptr<novaddt::HitList> slice_ptr(slices, i_slice);
            
            // get the Ptr for the trigger decision
            art::ProductID td_id = this->getProductID<std::vector<novaddt::TriggerDecision> >(event);
            art::Ptr<novaddt::TriggerDecision> td_ptr(td_id, trigger_decisions->size()-1, event.productGetter(td_id));
    
            // Association 
            assn->addSingle(td_ptr, slice_ptr);
            
            if (_verbose) std::cout<< "\t\t - passed the trigger decision, start time: "
                      << trigger_decisions->back().start()
                      << ", duration: " << trigger_decisions->back().duration()
                      << std::endl;
        } // end of check on prescale
        
    } // end of loop on tracks
    
    bool goodTrigger = (trigger_decisions->size() > 0);
    if (_verbose) std::cout << "=== end of novaddt::NuMuTrigger filter. Trigger decision: " << goodTrigger
              << " from " << trigger_decisions->size()
              << " passes (associations: " << assn->size()
              << ")" << std::endl;
              
    event.put(std::move(trigger_decisions));
    event.put(std::move(assn));
        
    return goodTrigger;
}
//---------------------------------------------------------------------
void novaddt::NuMuTrigger::findBoundingBox(novaddt::HitList & xz_hits, novaddt::HitList & yz_hits, 
                                                TVector3 & start, TVector3 & end)
{
    //if (_verbose) std::cout << "finding bounding box, " 
                            //<< xz_hits.size() << " xz hits, "
                            //<< yz_hits.size() << " yz hits"
                            //<< std::endl;
    // find z boundaries
    // default to the mid point
    int xz_z_min = 448;
    int xz_z_max = 448;
    int xz_x_min = 192;
    int xz_x_max = 192;
    if (xz_hits.size() > 0){
        xz_z_min = xz_hits[0].Plane().val;
        xz_z_max = xz_hits[0].Plane().val;
        xz_x_min = xz_hits[0].Cell().val;
        xz_x_max = xz_hits[0].Cell().val;
    }
    for (auto const& hit : xz_hits){
        if (hit.Plane().val < xz_z_min) xz_z_min = hit.Plane().val;
        if (hit.Plane().val > xz_z_max) xz_z_max = hit.Plane().val;
        if (hit.Cell().val  < xz_x_min) xz_x_min = hit.Cell().val;
        if (hit.Cell().val  > xz_x_max) xz_x_max = hit.Cell().val;
    }
    
    int yz_z_min = 448;
    int yz_z_max = 448;
    int yz_y_min = 192;
    int yz_y_max = 192;
    if (yz_hits.size() > 0){
        yz_z_min = yz_hits[0].Plane().val;
        yz_z_max = yz_hits[0].Plane().val;
        yz_y_min = yz_hits[0].Cell().val;
        yz_y_max = yz_hits[0].Cell().val;
    }
    for (auto const& hit : yz_hits){
        if (hit.Plane().val < yz_z_min) yz_z_min = hit.Plane().val;
        if (hit.Plane().val > yz_z_max) yz_z_max = hit.Plane().val;
        if (hit.Cell().val  < yz_y_min) yz_y_min = hit.Cell().val;
        if (hit.Cell().val  > yz_y_max) yz_y_max = hit.Cell().val;
    }
    
    int z_min = std::min(xz_z_min,yz_z_min);
    int z_max = std::max(xz_z_max,yz_z_max);
    
    start.SetXYZ( double (xz_x_min),
                  double (yz_y_min),
                  double (z_min));
    end.SetXYZ(   double (xz_x_max),
                  double (yz_y_max),
                  double (z_max));
    return;
}
//---------------------------------------------------------------------
double novaddt::NuMuTrigger::n3DCellsToEdge(novaddt::Track3D & track, bool entering)
{
    //std::cout<< "--- Find distance to Edge, entering: " << entering << std::endl;
    //// Debug
    //std::cout<< "--- track xz: " << xz_track.Start().X()
              //<< " - "            << xz_track.End().X()
              //<< ", z: "          << xz_track.Start().Z()
              //<< " - "            << xz_track.End().Z()
              //<< std::endl;
              
    //std::cout<< "--- track yz: " << yz_track.Start().X()
              //<< " - "            << yz_track.End().X()
              //<< ", z: "          << yz_track.Start().Z()
              //<< " - "            << yz_track.End().Z()
              //<< std::endl;

    // get the direction of the track
    // x
    double min_X = std::min(track.Start().X(), track.End().X());
    double max_X = std::max(track.Start().X(), track.End().X());
    // y
    double min_Y = std::min(track.Start().Y(), track.End().Y());
    double max_Y = std::max(track.Start().Y(), track.End().Y());
    // Z
    double min_Z = std::min(track.Start().Z(), track.End().Z());
    double max_Z = std::max(track.Start().Z(), track.End().Z());
    // direction
    TVector3 trajectory = TVector3(max_X - min_X,
                                   max_Y - min_Y,
                                   max_Z - min_Z);
    TVector3 dir = trajectory.Unit();
    
    double xlow, xhigh, ylow, yhigh, zlow, zhigh;
    if (_detector == "FD"){
        xlow  = 0;
        xhigh = 383;
        ylow  = 0;
        yhigh = 383;
        zlow  = 0;
        zhigh = 895;
    } else if (_detector == "FD_2DB"){
        xlow  = 0;
        xhigh = 383;
        ylow  = 0;
        yhigh = 383;
        zlow  = 0;
        zhigh = 127;
    } else {
        // default to NDOS
        xlow  = 0;
        xhigh = 63;
        ylow  = 0;
        yhigh = 95;
        zlow  = 0;
        zhigh = 198;
    }
    // Make sure we're inside the box!
    // if not set the track edge to match the detector boundary
    if (min_X<=xlow)  min_X = xlow;
    if (max_X>=xhigh) max_X = xhigh;
    if (min_Y<=ylow)  min_Y = ylow;
    if (max_Y>=yhigh) max_Y = yhigh;
    if (min_Z<=zlow)  min_Z = zlow;
    if (max_Z>=zhigh) max_Z = zhigh;  

    // If we're looking for entering tracks, then we don't want to check containment to the lower Z edge:
    if (entering) zlow = -999;
    
    // Compute the distances to the x/y/z walls
    double dx = 99.E99;
    double dy = 99.E99;
    double dz = 99.E99;
    // we will assume that the track could be going in either direction
    dx = std::min( ((xhigh - max_X)/dir.X()),
                   ((min_X - xlow) /dir.X())
                   );
    dy = std::min( ((yhigh - max_Y)/dir.Y()),
                   ((min_Y - ylow) /dir.Y())
                   );
    dz = std::min( ((zhigh - max_Z)/dir.Z()),
                   std::fabs((min_Z - zlow) /dir.Z()) // introduce fabs here so that we can offset the z_low for entering tracks
                   );
    // Choose the shortest distance
    double d = 0.0;
    if      (dx<dy && dx<dz) d = dx;
    else if (dy<dz && dy<dx) d = dy;
    else if (dz<dx && dz<dy) d = dz;

    // Make the step
    TVector3 distance = TVector3(dir.X()*d, dir.Y()*d, dir.Z()*d);
    
    //// Debug
    //if (_verbose) std::cout<< "--- combined track x: " << min_X
              //<< " - "                    << max_X
              //<< ", y: "                  << min_Y
              //<< " - "                    << max_Y
              //<< ", z: "                  << min_Z
              //<< " - "                    << max_Z
              //<< std::endl;
              
    //if (_verbose) std::cout<< "--- unit dir x: " << dir.X()
              //<< ", y: "            << dir.Y()
              //<< ", z: "            << dir.Z()
              //<< std::endl;
              
    //if (_verbose) std::cout<< "--- dx: " << dx
              //<< ", dy: "   << dy
              //<< ", dz: "   << dz
              //<< " => d: "  << d
              //<< std::endl;

    //if (_verbose) std::cout<< "--- distance x: " << distance.X()
              //<< ", y: "            << distance.Y()
              //<< ", z: "            << distance.Z()
              //<< ", mag: "          << distance.Mag()
              //<< std::endl;
    //assert(distance.Mag() > 1.);
    return distance.Mag();
}
//---------------------------------------------------------------------
void novaddt::NuMuTrigger::endJob()
{
    std::cout << "--- novaddt::NuMuTrigger endJob" << std::endl;
    std::cout << "\tNumber of slices seen:          " << _nSlices             << std::endl;
    std::cout << "\t - pass slice containment:      " << _nPass_slice_containment << std::endl;
    std::cout << "\tNumber of tracks seen:          " << _nTracks             << std::endl;
    std::cout << "\tNumber of 3D tracks seen:       " << _n3DTracks           << std::endl;
    std::cout << "\t - have a 3D track:             " << _nPass_3DTracks      << std::endl;
    std::cout << "\t - pass dX:                     " << _nPass_dX            << std::endl;
    std::cout << "\t - pass dY:                     " << _nPass_dY            << std::endl;
    std::cout << "\t - pass dZ:                     " << _nPass_dZ            << std::endl;
    std::cout << "\t - pass projection containment: " << _nPass_projection_containment  << std::endl;
    std::cout << "\t - pass box containment:        " << _nPass_box_containment         << std::endl;
    std::cout << "\t - pass angle:                  " << _nPass_angle         << std::endl;
    std::cout << "\t - pass entering:               " << _nPass_entering      << std::endl;
    std::cout << "\t - pass cosmic PID:             " << _nPass_cosmic_PID    << std::endl;
    std::cout << "\t - pass angle variable:         " << _nPass_angle_var     << std::endl;
    std::cout << "\t - pass the trigger:            " << _trigger_counts      << std::endl;
    std::cout << "\t              after prescale:   " << _prescaled_trigger_counts << std::endl;
}
//---------------------------------------------------------------------
DEFINE_ART_MODULE(novaddt::NuMuTrigger)
