////////////////////////////////////////////////////////////////////////
// Class:       Merge2DTracks
// Module Type: filter
// File:        Merge2DTracks_module.cc
//
// Generated at Tue Jun 25 06:01:01 2013 by Matthew Tamsett using artmod
// from cetpkgsupport v1_02_00.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Persistency/Common/Assns.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/Track.h"
#include "DDTBaseDataProducts/Track3D.h"
#include "DDTBaseDataProducts/HitList.h"

#include "DAQChannelMap/DAQChannelMap.h"
#include "DDTUtilities/AssociationUtil.h"

#include <unordered_set>
#include <memory>
#include <vector>
//------------------------------------------------------------------------
namespace novaddt {
    class Merge2DTracks;

    const float DELTA = 1e-5;
    const float DELTA_TDC = 1e9; 

    inline bool same_track(const Track *lhs, const Track *rhs) {
        return lhs == rhs;
                /*(fabs(lhs.StartT()-rhs.StartT())<DELTA_TDC &&
                fabs(lhs.EndT()-rhs.EndT())    <DELTA_TDC &&
                fabs(lhs.StartV()-rhs.StartV())<DELTA     &&
                fabs(lhs.StartZ()-rhs.StartZ())<DELTA     &&
                fabs(lhs.EndV()-rhs.EndV())    <DELTA     &&
                fabs(lhs.EndZ()-rhs.EndZ())    <DELTA     &&
                lhs.View()   == rhs.View());*/
    }
}
//------------------------------------------------------------------------
class novaddt::Merge2DTracks : public art::EDFilter {
    public:
        explicit Merge2DTracks(fhicl::ParameterSet const & p);
        virtual ~Merge2DTracks();
        bool    filter(art::Event & e) override;
        void    endJob() override;
        void    fillDummyTracks(std::vector< art::Ptr<novaddt::Track> > tracks, std::unique_ptr< std::vector<novaddt::Track3D>   > & merged_tracks);
        void    simpleMerge(std::vector< art::Ptr<novaddt::Track> > tracks, 
                            art::Handle< std::vector<novaddt::Track> > all_tracks,
                            std::unique_ptr< std::vector<novaddt::Track3D> > & merged_tracks, 
                            std::unique_ptr< std::vector<novaddt::HitList> > & hitListcol,
                            art::FindOneP<novaddt::HitList> fohl);
        void    twoViewMerge(std::vector< art::Ptr<novaddt::Track> > tracks,
                             art::Handle< std::vector<novaddt::Track> > all_tracks,
                             std::unique_ptr< std::vector<novaddt::Track3D> > & merged_tracks, 
                             std::unique_ptr< std::vector<novaddt::HitList> > & hitListcol,
                             art::FindOneP<novaddt::HitList> fohl);
        double  lengthOfTrackPair(art::Ptr<novaddt::Track> x_track, art::Ptr<novaddt::Track> y_track);
        double  nHitsOfTrackPair(art::FindOneP<novaddt::HitList> fohl, unsigned int i, unsigned int j);
    private:
        std::string  _trackModuleLabel;         ///< label of module making the Tracks
        std::string  _trackInstanceLabel;       ///< instance label making the Tracks
        std::string  _sliceModuleLabel;         ///< label of module making the HitList
        std::string  _sliceInstanceLabel;       ///< instance label making the HitList
        std::string  _trackToSliceInstanceLabel;///< instance of track to slice association produced by track module
        bool         _simpleTwoViewMode;        ///< A boolean switch to a mode where we expect exactly two tracks and we combine these without checking metrics
        bool         _alwaysMergeMin;           ///< A boolean switch to force the merge of the track pair with the minimum metric
        std::string  _matchingMetric;           ///< the method used to match tracks
                                                ///< currently configured options are: 
                                                ///<        dZ - sqrt(dZ_start^2 + dZ_end^2)
        std::string  _ambiguityMetric;          ///< the method used to break matching degeneracies
                                                ///< currently configured options are: 
                                                ///<       length - the longest pair are selected
                                                ///<       nhits - the pair with the most hits are selected
        double       _metricThreshold;          ///< the maximum value of metric which can be considered matched
        unsigned int _minhits;                  ///< the maximum value of metric which can be considered matched
        int          _nSlices = 0;              ///< number of slices seen
        int          _nTracks = 0;              ///< number of tracks seen
        int          _nMergedTracks_3D = 0;     ///< number of merged tracks produced that are 3D
        int          _nMergedTracks_2D = 0;     ///< number of merged tracks produced that are 2D
        int          _nMergedTracks = 0;        ///< number of merged tracks produced
     
        std::unordered_set<size_t> unseenTracks;   
};
//------------------------------------------------------------------------
novaddt::Merge2DTracks::Merge2DTracks(fhicl::ParameterSet const & p)
:   _trackModuleLabel           (p.get<std::string >("TrackModuleLabel"      ))
  , _trackInstanceLabel         (p.get<std::string >("TrackInstanceLabel", ""))
  , _sliceModuleLabel           (p.get<std::string >("SliceModuleLabel"      ))
  , _sliceInstanceLabel         (p.get<std::string >("SliceInstanceLabel", ""))
  , _trackToSliceInstanceLabel  (p.get<std::string >("TrackToSliceInstanceLabel", ""))
  , _simpleTwoViewMode          (p.get<bool>        ("SimpleTwoViewMode"     ))
  , _alwaysMergeMin             (p.get<bool>        ("AlwaysMergeMin"        ))
  , _matchingMetric             (p.get<std::string >("MatchingMetric",     ""))
  , _ambiguityMetric            (p.get<std::string >("AmbiguityMetric",    ""))
  , _metricThreshold            (p.get<double>      ("MetricThreshold"       ))
  , _minhits                    (p.get<int>         ("MinHits"               ))
{
    std::cout << "=== novaddt::Merge2DTracks instantiate" << std::endl;
    std::cout << "\t TrackModuleLabel:          " << _trackModuleLabel          << std::endl;
    std::cout << "\t TrackInstanceLabel:        " << _trackInstanceLabel        << std::endl;
    std::cout << "\t SliceModuleLabel:          " << _sliceModuleLabel          << std::endl;
    std::cout << "\t SliceInstanceLabel:        " << _sliceInstanceLabel        << std::endl;
    std::cout << "\t TrackToSliceInstanceLabel: " << _trackToSliceInstanceLabel << std::endl;
    std::cout << "\t SimpleTwoViewMode:         " << _simpleTwoViewMode         << std::endl;
    std::cout << "\t AlwaysMergeMin:            " << _alwaysMergeMin            << std::endl;
    std::cout << "\t MatchingMetric:            " << _matchingMetric            << std::endl;
    std::cout << "\t AmbiguityMetric:           " << _ambiguityMetric           << std::endl;
    std::cout << "\t MetricThreshold:           " << _metricThreshold           << std::endl;
    
    // check that the requested matching metric is configured
    assert( (_matchingMetric == "dZ") ); 
    // check that the requested ambiguity metric is configured
    assert( (_ambiguityMetric == "length") ||
            (_ambiguityMetric == "nhits")
          );

    produces< std::vector<novaddt::Track3D> >();
    produces< std::vector<novaddt::HitList> >();
    
    produces< art::Assns<novaddt::Track3D, novaddt::HitList> >();
    produces< art::Assns<novaddt::Track3D, novaddt::HitList> >(_trackToSliceInstanceLabel);
}
//------------------------------------------------------------------------
novaddt::Merge2DTracks::~Merge2DTracks()
{
    // Clean up dynamic memory and other resources here.
}
//------------------------------------------------------------------------
bool novaddt::Merge2DTracks::filter(art::Event & event)
{
    // Implementation of required member function here.
    //LOG_DEBUG("Merge2DTracks") << "=== novaddt::Merge2DTracks filter. Event: " 
    //          << event.id().event() 
    //          << std::endl;
    
    // make the final data products
    std::unique_ptr< std::vector<novaddt::Track3D> > merged_tracks    (new std::vector<novaddt::Track3D>);
    std::unique_ptr< std::vector<novaddt::HitList> > merged_hit_lists (new std::vector<novaddt::HitList>);
    std::unique_ptr< art::Assns<novaddt::Track3D,   novaddt::HitList> > assn(new art::Assns<novaddt::Track3D, novaddt::HitList>);
    std::unique_ptr< art::Assns<novaddt::Track3D,   novaddt::HitList> > assn_b(new art::Assns<novaddt::Track3D, novaddt::HitList>);

    // get the slices
    art::Handle< std::vector<novaddt::HitList> > slices;
    event.getByLabel(_sliceModuleLabel, _sliceInstanceLabel, slices);
    //LOG_DEBUG("Merge2DTracks") << "\tgot " << slices->size() << " slices" << std::endl;
    _nSlices+=slices->size();
    
    //Now make it an art::PtrVector so we can use associations later
    art::PtrVector<novaddt::HitList> slicelist;
    for(unsigned int i = 0; i < slices->size(); ++i){
        art::Ptr<novaddt::HitList>slice(slices,i);
        slicelist.push_back(slice);
    }
    
    // ---- start of bug fix ----
    // Due to a bug in ART v1_08_09 in order to use the one to many association later
    // we first have to access the one to one.
    // This finder can be used later so it's not all a waste of time
    // in fact this might speed up the algorithm!
    art::Handle< std::vector<novaddt::Track> > all_tracks;
    event.getByLabel(_trackModuleLabel, _trackInstanceLabel, all_tracks);
    //// get the hit list for each track
    art::FindOneP<novaddt::HitList> fohl(all_tracks, event, _trackModuleLabel);
    //assert(fohl.isValid());
    // ---- end of bug fix ----
    
    

    // populate unseenTracks set:
    unseenTracks.clear();
    for (size_t i=0; i<all_tracks->size(); ++i) unseenTracks.insert(i);


    const std::string the_slices = _trackModuleLabel+":"+_trackToSliceInstanceLabel;
    // get the tracks for each hit list
    //art::FindManyP<novaddt::Track> find_tracks(slices, event, _trackModuleLabel);
    art::FindManyP<novaddt::Track> find_tracks(slices, event, the_slices);
    assert(find_tracks.isValid());
    
    // loop over slices
    for(size_t i_slice = 0; i_slice < slices->size(); ++i_slice){
        // get the hit list
        novaddt::HitList hit_list = slices->at(i_slice);
	if(hit_list.size() < _minhits) continue;
        //LOG_DEBUG("Merge2DTracks") << "\tslice[" << i_slice << "]: " << hit_list.size() << " hits" << std::endl;
        // loop over hits, useful for debugging only at the moment
        //for(size_t i_hit = 0; i_hit < hit_list.size(); ++i_hit){
            //novaddt::DAQHit hit(hit_list[i_hit]);
            //std::string view = "y";
            //if( hit.View().val == daqchannelmap::X_VIEW ) view = "x";
            //LOG_DEBUG("Merge2DTracks") << "\t\thit[" << i_hit
                      //<< "]: TDC: "   << hit.TDC().val
                      //<< ", ADC: "    << hit.ADC().val
                      //<< ", plane: "  << hit.Plane().val
                      //<< ", "         << view
                      //<< "-cell: "    << hit.Cell().val
                      //<< std::endl;         
        //} // end of loop on hits
        // find the tracks for this hit list
        std::vector< art::Ptr<novaddt::Track> > tracks = find_tracks.at(i_slice);


        //LOG_DEBUG("Merge2DTracks") << "\t - got " << tracks.size() << " tracks" << std::endl;
        // if there are no tracks we continue
        if (!tracks.size()) continue;
        _nTracks+=tracks.size();
        
        // if simple two view mode then we must get exactly two tracks
        assert(!_simpleTwoViewMode || (tracks.size()==2));
        
        // get the hit list for each track
        // removed in order to improve speed, see Merge functions
        //art::FindOneP<novaddt::HitList> fohl(tracks, event, _trackModuleLabel);
        //assert(fohl.isValid());
        //assert(fohl.size() == tracks.size());
    
        //for(size_t i_track = 0; i_track < tracks.size(); ++i_track){
            //art::Ptr<novaddt::Track> track = tracks[i_track];
            //std::string view = "y";
            //if( track->View() == daqchannelmap::X_VIEW ) view = "x";
            //// get the hit list for this track
            //art::Ptr<novaddt::HitList> this_hit_list = fohl.at(i_track);
            
            ////LOG_DEBUG("Merge2DTracks") << "\t\ttrack[" << i_track
                      ////<< "]\t view: " << view
                      ////<< ", hits: "   << this_hit_list->size() 
                      //////<< ", time: "   << track->StartT() << " - " << track->EndT()
                      ////<< ", V: "      << track->StartV() << " - " << track->EndV()
                      ////<< ", Z: "      << track->StartZ() << " - " << track->EndZ()
                      ////<< std::endl;
                      
            ////// loop over hits, useful for debugging only at the moment
            ////for(size_t i_hit = 0; i_hit < this_hit_list->size(); ++i_hit){
                ////std::string h_view = "y";
                ////if( this_hit_list->at(i_hit).View().val == daqchannelmap::X_VIEW) h_view = "x";
                ////LOG_DEBUG("Merge2DTracks") << "\t\t\thit["   << i_hit
                          ////<< "]: \tplane: " << this_hit_list->at(i_hit).Plane().val
                          ////<< ",\tcell: "    << this_hit_list->at(i_hit).Cell().val
                          ////<< ",\t view: "   << h_view
                          ////<< ",\tTDC: "     << this_hit_list->at(i_hit).TDC().val
                          ////<< ",\tADC: "     << this_hit_list->at(i_hit).ADC().val
                          ////<< std::endl;
            ////} // end of loop on hits
        //} // end of loop on tracks
        
        // merge tracks
        //fillDummyTracks(tracks, merged_tracks);
        unsigned int current_size = merged_tracks->size();
        if (_simpleTwoViewMode){
            twoViewMerge(tracks, all_tracks, merged_tracks, merged_hit_lists, fohl);
        } else {
            simpleMerge(tracks, all_tracks, merged_tracks, merged_hit_lists, fohl);
        }
        
        //LOG_DEBUG("Merge2DTracks") << "\t - produced " << merged_tracks->size()    << " tracks" << std::endl;
        //LOG_DEBUG("Merge2DTracks") << "\t - produced " << merged_hit_lists->size() << " hit lists" << std::endl;
        
        // the number of merged tracks should ALWAYS be equal to the number of hit lists
        assert(merged_tracks->size()==merged_hit_lists->size());
        
        //LOG_DEBUG("Merge2DTracks") << "\t - saving associations" << std::endl;
        // loop over new tracks and hit lists, save final products and associations
        for (unsigned int i = current_size; i < merged_tracks->size(); ++i){
            //LOG_DEBUG("Merge2DTracks") << "\t\tassociating track[" << i << "] to slice[" << i_slice << "]" << std::endl;
            // association between this slice and each track
            util::CreateAssn(*this, event, *merged_tracks, slicelist[i_slice], *assn_b, i);
            // association between each track and it's hits
            //LOG_DEBUG("Merge2DTracks") << "\t\tassociating track[" << i << "] to hit list[" << i << " - " << i+1 << "]" << std::endl;
            util::CreateAssn(*this, event, *merged_tracks, *merged_hit_lists, *assn, i, i+1, i);
        }
        
    } // end of loop on slices
    
    // store the merged tracks
    _nMergedTracks += merged_tracks->size();
    event.put(std::move(merged_tracks));
    event.put(std::move(merged_hit_lists));
    event.put(std::move(assn));
    event.put(std::move(assn_b), _trackToSliceInstanceLabel);
    return true;
}
//------------------------------------------------------------------------
void novaddt::Merge2DTracks::fillDummyTracks(std::vector< art::Ptr<novaddt::Track> > tracks,
                                             std::unique_ptr< std::vector<novaddt::Track3D>   > & merged_tracks)
{
    //LOG_DEBUG("Merge2DTracks") << "\tFilling dummy tracks" << std::endl;
    for(size_t i_track = 0; i_track < tracks.size(); ++i_track){
        art::Ptr<novaddt::Track> track = tracks[i_track];
        int view = 0;
        TVector3 start;
        TVector3 end;
        
        if( track->View() == daqchannelmap::X_VIEW ){
            view = 1;
            start.SetXYZ(track->StartV(), 0., track->StartZ());
            end.SetXYZ(  track->EndV(),   0., track->EndZ());
        } else {
            start.SetXYZ(0., track->StartV(), track->StartZ());
            end.SetXYZ(  0., track->EndV(),   track->EndZ());
        }
        
        novaddt::Track3D merged_track = Track3D(false, view, start, end);
        
        //LOG_DEBUG("Merge2DTracks") << "\t\tmerged track[" << i_track
                      //<< "]\t Is3D: "    << merged_track.Is3D()
                      //<< ", view: "      << merged_track.View()
                      //<< ", X: "         << merged_track.Start().X() << " - " << merged_track.End().X()
                      //<< ", Y: "         << merged_track.Start().Y() << " - " << merged_track.End().Y()
                      //<< ", Z: "         << merged_track.Start().Z() << " - " << merged_track.End().Z()
                      //<< std::endl;
                      
        merged_tracks->push_back(merged_track);
    } // end of loop on tracks
}
//------------------------------------------------------------------------
void novaddt::Merge2DTracks::simpleMerge(std::vector< art::Ptr<novaddt::Track> > tracks,
                                         art::Handle< std::vector<novaddt::Track> > all_tracks,
                                         std::unique_ptr< std::vector<novaddt::Track3D> > & merged_tracks,
                                         std::unique_ptr< std::vector<novaddt::HitList> > & merged_hit_lists,
                                         art::FindOneP<novaddt::HitList> fohl)
{
    //LOG_DEBUG("Merge2DTracks") << "\tMerging tracks" << std::endl;
    int nXTracks = 0;
    int nYTracks = 0;
    for(size_t i_track = 0; i_track < tracks.size(); ++i_track){
        if (tracks[i_track]->View() == daqchannelmap::X_VIEW){
            nXTracks++;
        } else {
            nYTracks++;
        }
    }
    
    //LOG_DEBUG("Merge2DTracks") << "\t - " << nXTracks << " x-tracks" << std::endl;
    //LOG_DEBUG("Merge2DTracks") << "\t - " << nYTracks << " y-tracks" << std::endl;
    
    if ( (nXTracks == 0) && (nYTracks == 0) ) return;
    
    //LOG_DEBUG("Merge2DTracks") << "\t - make metric matrix" << std::endl;
    // construct metric matrix
    double dummy_metric = 1e9;
    double metric_matrix[tracks.size()][tracks.size()];
    // fill with large dummy values
    for (unsigned int i = 0; i < tracks.size(); ++i){
        for (unsigned int j = 0; j < tracks.size(); ++j){
            metric_matrix[i][j] = dummy_metric;
        }
    }

    // for each track in slice, find the index into fohl that we must use
    std::vector<size_t> fohl_index;

    for (size_t i_track = 0; i_track < tracks.size(); ++i_track) {
      for (auto j_it = unseenTracks.begin(); j_it != unseenTracks.end(); ++j_it)
      {
        // this shouldn't happen, but here for safety
        if (all_tracks->size() <= *j_it) continue;

        if (same_track(&(*tracks.at(i_track)), &all_tracks->at(*j_it))) {
           fohl_index.push_back(*j_it);//assign(i_track, j_track);
           // erase track at position j_track
           unseenTracks.erase(j_it);
           break;
        }
      }
    }

    if (fohl_index.size() != tracks.size()) return; // can't match all tracks, bail


    // double loop over tracks to fill metric matrix
    //LOG_DEBUG("Merge2DTracks") << "\t - fill metric matrix" << std::endl;
    for(size_t i_track = 0; i_track < tracks.size(); ++i_track){
        //LOG_DEBUG("Merge2DTracks") << "\t - x-track[" << i_track << "]" << std::endl;
        if (tracks[i_track]->View() != daqchannelmap::X_VIEW) continue;
        art::Ptr<novaddt::Track> x_track = tracks[i_track];
        
        for(size_t j_track = 0; j_track < tracks.size(); ++j_track){
            //LOG_DEBUG("Merge2DTracks") << "\t\t - y-track[" << j_track << "]" << std::endl;
            if (tracks[j_track]->View() == daqchannelmap::X_VIEW) continue;
            art::Ptr<novaddt::Track> y_track = tracks[j_track];
            
            // define the distance measure
            // this could be anything we want
            // start with sqrt(dZ_low^2 + dZ_high^2)
            double x_low_z  = std::min(x_track->StartZ(),x_track->EndZ());
            double x_high_z = std::max(x_track->StartZ(),x_track->EndZ());
            double y_low_z  = std::min(y_track->StartZ(),y_track->EndZ());
            double y_high_z = std::max(y_track->StartZ(),y_track->EndZ());
            
            double dZ_low  = x_low_z  - y_low_z;
            double dZ_high = x_high_z - y_high_z;
            
            if (_matchingMetric=="dZ") metric_matrix[i_track][j_track] = std::sqrt( (dZ_low*dZ_low) + (dZ_high*dZ_high) );
            
            //LOG_DEBUG("Merge2DTracks") << "\t\tmetric: "      << _matchingMetric
                      //<< " between x-track[" << i_track
                      //<< "] and y-track["    << j_track
                      //<< "]: "               << metric_matrix[i_track][j_track] 
                      //<< std::endl;
        } // end of loop on y
    } // end of loop on x
    
    // now create merged tracks for those for which the metric is below threshold
    int iteration_counter = 0;
    int n_merged_tracks   = 0;
    std::vector<unsigned int> used_tracks;
    
    while (true){
        iteration_counter++;
        //LOG_DEBUG("Merge2DTracks") << "iteration: " << iteration_counter << std::endl;
        double min_metric = 9e9;
        int min_x = 1e9;
        size_t min_x_index = 1e9;
        int min_y = 1e9;
        size_t min_y_index = 1e9;
        double ambiguity_metric = 1e9;
        bool any_min = false;
        
        // loop over x
        for (unsigned int i = 0; i < tracks.size(); ++i){
            if (tracks[i]->View() != daqchannelmap::X_VIEW) continue;
            // if used continue
            if(std::find(used_tracks.begin(), used_tracks.end(), i)!=used_tracks.end()){
                //LOG_DEBUG("Merge2DTracks") << "\tx-track[" << i << "]: used" << std::endl;
                continue;
            }

            size_t i_index = fohl_index.at(i);
            // loop over y
            for (unsigned int j = 0; j < tracks.size(); ++j){
                if (tracks[j]->View() == daqchannelmap::X_VIEW) continue;
                // if used continue
                if(std::find(used_tracks.begin(), used_tracks.end(), j)!=used_tracks.end()){
                    //LOG_DEBUG("Merge2DTracks") << "\ty-track[" << j << "]: used" << std::endl;
                    continue;
                }
                // get metric
                //LOG_DEBUG("Merge2DTracks") << "\t - metric between x-track[" << i
                          //<< "] and y-track[" << j
                          //<< "]: " << metric_matrix[i][j] <<std::endl;
                // if the value wasn't filled continue
                if (metric_matrix[i][j] == dummy_metric) continue;

                size_t j_index = fohl_index.at(j);

                // check if new minimum
                bool new_min = false;
                if (metric_matrix[i][j] < min_metric) new_min = true;
                // if identical pick the pair with the lowest ambiguity metric
                if (metric_matrix[i][j] == min_metric){
                    double this_ambiguity_metric = 0;
                    if (_ambiguityMetric=="length") this_ambiguity_metric = lengthOfTrackPair(tracks[i],tracks[j]);
                    if (_ambiguityMetric=="nhits")  this_ambiguity_metric = nHitsOfTrackPair(fohl,i_index,j_index);
                    //LOG_DEBUG("Merge2DTracks") << "\t\t - indentical metrics, picking based on " << _ambiguityMetric << ": " << this_ambiguity_metric << std::endl;
                    if (this_ambiguity_metric > ambiguity_metric) new_min = true;
                } // end of ambiguity solver
                
                if (new_min){
                    // save new min
                    min_metric = metric_matrix[i][j];
                    min_x = i;
                    min_x_index = i_index;
                    min_y = j;
                    min_y_index = j_index;
                    any_min = true;
                    if (_ambiguityMetric=="length") ambiguity_metric = lengthOfTrackPair(tracks[i],tracks[j]);
                    if (_ambiguityMetric=="nhits")  ambiguity_metric = nHitsOfTrackPair(fohl,i_index,j_index);
                                                       
                    //LOG_DEBUG("Merge2DTracks") << "\t\t - new min, " << _ambiguityMetric
                              //<< ": "               << ambiguity_metric << std::endl;
                }
            } // end of loop on y
        } // end of loop on x
        
        if ((min_metric<_metricThreshold) || \
            (_alwaysMergeMin && (n_merged_tracks == 0) && any_min)){
            n_merged_tracks++;
            //LOG_DEBUG("Merge2DTracks") << "Saving minimum: " << min_metric
                      //<< " from [" << min_x 
                      //<< ","       << min_y 
                      //<< "]"       << std::endl;
            used_tracks.push_back(min_x);
            used_tracks.push_back(min_y);
            
            // save merged track
            TVector3 start;
            TVector3 end;
            // assume north going, that is the min Z is the start point
            // find the minimum x-z
            // assume that track is already north going
            double start_x   = tracks[min_x]->StartV();
            double end_x     = tracks[min_x]->EndV();
            double min_xz_z  = tracks[min_x]->StartZ();
            double max_xz_z  = tracks[min_x]->EndZ();
            // but check for south going
            if (tracks[min_x]->StartZ() > tracks[min_x]->EndZ() ){
                start_x  = tracks[min_x]->EndV();
                end_x    = tracks[min_x]->StartV();
                min_xz_z = tracks[min_x]->EndZ();
                max_xz_z = tracks[min_x]->StartZ();
            }
            // repeat for y-x
            double start_y   = tracks[min_y]->StartV();
            double end_y     = tracks[min_y]->EndV();
            double min_yz_z  = tracks[min_y]->StartZ();
            double max_yz_z  = tracks[min_y]->EndZ();
            // but check for south going
            if (tracks[min_y]->StartZ() > tracks[min_y]->EndZ() ){
                start_y  = tracks[min_y]->EndV();
                end_y    = tracks[min_y]->StartV();
                min_yz_z = tracks[min_y]->EndZ();
                max_yz_z = tracks[min_y]->StartZ();
            }
            
            double start_z = std::min(min_xz_z, min_yz_z);
            double end_z   = std::max(max_xz_z, max_yz_z);
            
            start.SetXYZ(start_x,
                         start_y,
                         start_z);
            end.SetXYZ(  end_x,
                         end_y,
                         end_z);
                         
            novaddt::Track3D merged_track = Track3D(true, 2, start, end);
            
            //LOG_DEBUG("Merge2DTracks") << "\t\tmerged track:" 
                          //<< ",  Is3D: "     << merged_track.Is3D()
                          //<< ", view: "      << merged_track.View()
                          //<< ", X: "         << merged_track.Start().X() << " - " << merged_track.End().X()
                          //<< ", Y: "         << merged_track.Start().Y() << " - " << merged_track.End().Y()
                          //<< ", Z: "         << merged_track.Start().Z() << " - " << merged_track.End().Z()
                          //<< std::endl;
            merged_tracks->push_back(merged_track);    
            
            // save merged hit list
            novaddt::HitList merged_hit_list;
            // get the hit list for each track
            novaddt::HitList x_hit_list = *fohl.at(min_x_index);
            for (auto const& hit : x_hit_list ) {
                merged_hit_list.emplace_back(hit);
            }
            novaddt::HitList y_hit_list = *fohl.at(min_y_index);
            for (auto const& hit : y_hit_list ) {
                merged_hit_list.emplace_back(hit);
            }            
            merged_hit_lists->push_back(merged_hit_list);
            
            _nMergedTracks_3D++;        
        } else {
            //LOG_DEBUG("Merge2DTracks") << "Minimum: "      << min_metric
                      //<< " > threshold: " << _metricThreshold
                      //<< std::endl;
            break;
        }
    } // end of while   
    // loop over tracks and store the remaining ones as 2-D tracks
    for(unsigned int i_track = 0; i_track < tracks.size(); ++i_track){
        if(std::find(used_tracks.begin(), used_tracks.end(), i_track)!=used_tracks.end()) continue;
        art::Ptr<novaddt::Track> track = tracks[i_track];
        int view = 0;
        TVector3 start;
        TVector3 end;
        
        if( track->View() == daqchannelmap::X_VIEW ){
            view = 1;
            start.SetXYZ(track->StartV(), 0., track->StartZ());
            end.SetXYZ(  track->EndV(),   0., track->EndZ());
        } else {
            start.SetXYZ(0., track->StartV(), track->StartZ());
            end.SetXYZ(  0., track->EndV(),   track->EndZ());
        }
        
        novaddt::Track3D merged_track = Track3D(false, view, start, end);
        
        //LOG_DEBUG("Merge2DTracks") << "\t\tmerged (2D) track[" << i_track
                      //<< "]\t Is3D: "    << merged_track.Is3D()
                      //<< ", view: "      << merged_track.View()
                      //<< ", X: "         << merged_track.Start().X() << " - " << merged_track.End().X()
                      //<< ", Y: "         << merged_track.Start().Y() << " - " << merged_track.End().Y()
                      //<< ", Z: "         << merged_track.Start().Z() << " - " << merged_track.End().Z()
                      //<< std::endl;
                      
        merged_tracks->push_back(merged_track);
        
        // save merged hit list
        novaddt::HitList merged_hit_list;
        // get the hit list for the track
        novaddt::HitList hit_list = *fohl.at(fohl_index[i_track]);
        for (auto const& hit : hit_list ) {
            merged_hit_list.emplace_back(hit);
        }
        merged_hit_lists->push_back(merged_hit_list);
            
        _nMergedTracks_2D++; 
    } // end of loop on tracks
    //assert(false);
}
//------------------------------------------------------------------------
void novaddt::Merge2DTracks::twoViewMerge(std::vector< art::Ptr<novaddt::Track> > tracks,
                                          art::Handle< std::vector<novaddt::Track> > all_tracks,
                                          std::unique_ptr< std::vector<novaddt::Track3D> > & merged_tracks,
                                          std::unique_ptr< std::vector<novaddt::HitList> > & merged_hit_lists,
                                          art::FindOneP<novaddt::HitList> fohl)
{
    //LOG_DEBUG("Merge2DTracks") << "\tMerging tracks" << std::endl;
    // this method assumes that the Tracking module or an equivalent has made these tracks
    // such tracks always come in pairs and the x track is always the first
    // the fact that there is two has been asserted previously. 
    assert(tracks[0]->View() == daqchannelmap::X_VIEW);
    assert(tracks[1]->View() != daqchannelmap::X_VIEW);
    
    // save merged track
    TVector3 start;
    TVector3 end;
    // assume north going, that is the min Z is the start point
    // find the minimum x-z
    // assume that track is already north going
    double start_x   = tracks[0]->StartV();
    double end_x     = tracks[0]->EndV();
    double min_xz_z  = tracks[0]->StartZ();
    double max_xz_z  = tracks[0]->EndZ();
    // but check for south going
    if (tracks[0]->StartZ() > tracks[0]->EndZ() ){
        start_x  = tracks[0]->EndV();
        end_x    = tracks[0]->StartV();
        min_xz_z = tracks[0]->EndZ();
        max_xz_z = tracks[0]->StartZ();
    }
    // repeat for y-x
    double start_y   = tracks[1]->StartV();
    double end_y     = tracks[1]->EndV();
    double min_yz_z  = tracks[1]->StartZ();
    double max_yz_z  = tracks[1]->EndZ();
    // but check for south going
    if (tracks[1]->StartZ() > tracks[1]->EndZ() ){
        start_y  = tracks[1]->EndV();
        end_y    = tracks[1]->StartV();
        min_yz_z = tracks[1]->EndZ();
        max_yz_z = tracks[1]->StartZ();
    }
    
    size_t zero_index, one_index;
    bool zero_found = false, one_found = false;
    for (size_t j=0; j<all_tracks->size(); ++j)
    {
       if (!zero_found && same_track(&(*tracks[0]),&all_tracks->at(j))) {
          zero_index = j;
          zero_found = true;
       }
       if (!one_found && same_track(&(*tracks[1]),&all_tracks->at(j))) {
          one_index = j;
          one_found = true;
       }
       if (zero_found && one_found) break;
    }

    novaddt::HitList x_hit_list = *fohl.at(zero_index);
    novaddt::HitList y_hit_list = *fohl.at(one_index);
    double start_z,end_z;
    // If the x hit list is empty set z from to y
    if (!x_hit_list.size()){
        start_z = min_yz_z;
        end_z   = max_yz_z;
    // else if the y hit list is empty set z from to x
    } else if (!y_hit_list.size()){
        start_z = min_xz_z;
        end_z   = max_xz_z;
    // else use the minimum and maximum of both
    } else {
        start_z = std::min(min_xz_z, min_yz_z);
        end_z   = std::max(max_xz_z, max_yz_z);
    }
    
    start.SetXYZ(start_x,
                 start_y,
                 start_z);
    end.SetXYZ(  end_x,
                 end_y,
                 end_z);
                 
    novaddt::Track3D merged_track = Track3D(true, 2, start, end);
    
    //LOG_DEBUG("Merge2DTracks") << "\t\tmerged track:" 
                  //<< ",  Is3D: "     << merged_track.Is3D()
                  //<< ", view: "      << merged_track.View()
                  //<< ", X: "         << merged_track.Start().X() << " - " << merged_track.End().X()
                  //<< ", Y: "         << merged_track.Start().Y() << " - " << merged_track.End().Y()
                  //<< ", Z: "         << merged_track.Start().Z() << " - " << merged_track.End().Z()
                  //<< std::endl;
    merged_tracks->push_back(merged_track);    
    
    // save merged hit list
    novaddt::HitList merged_hit_list;
    // get the hit list for each track
    
    for (auto const& hit : x_hit_list ) {
        merged_hit_list.emplace_back(hit);
    }
    for (auto const& hit : y_hit_list ) {
        merged_hit_list.emplace_back(hit);
    }            
    merged_hit_lists->push_back(merged_hit_list);
    
    _nMergedTracks_3D++;
}
//------------------------------------------------------------------------
double novaddt::Merge2DTracks::lengthOfTrackPair(art::Ptr<novaddt::Track> x_track, art::Ptr<novaddt::Track> y_track)
{
    double this_dX  = std::fabs(x_track->EndV() - x_track->StartV());
    double this_dXZ = std::fabs(x_track->EndZ() - x_track->StartZ());
    double this_dY  = std::fabs(y_track->EndV() - y_track->StartV());
    double this_dYZ = std::fabs(y_track->EndZ() - y_track->StartZ());
    double length   = std::sqrt( (this_dX*this_dX)   +
                                 (this_dXZ*this_dXZ) +
                                 (this_dY*this_dY)   +
                                 (this_dYZ*this_dYZ) );
    return length;
}
//------------------------------------------------------------------------
double novaddt::Merge2DTracks::nHitsOfTrackPair(art::FindOneP<novaddt::HitList> fohl, unsigned int i, unsigned int j)
{
    double n_hits = 0;
    
    // get the hit list for each track
    art::Ptr<novaddt::HitList> x_hit_list = fohl.at(i);
    art::Ptr<novaddt::HitList> y_hit_list = fohl.at(j);
    
    n_hits+=double(x_hit_list->size());
    n_hits+=double(y_hit_list->size());
    
    return n_hits;
}
//------------------------------------------------------------------------
void novaddt::Merge2DTracks::endJob()
{
    std::cout << "--- novaddt::Merge2DTracks endJob " << std::endl;
    std::cout << "\tNumber of slices seen:          " << _nSlices           << std::endl;
    std::cout << "\tNumber of tracks seen:          " << _nTracks           << std::endl;
    std::cout << "\tNumber of merged tracks made:   " << _nMergedTracks     << std::endl;
    std::cout << "\tNumber of merged 3D tracks:     " << _nMergedTracks_3D  << std::endl;
    std::cout << "\tNumber of merged 2D tracks:     " << _nMergedTracks_2D  << std::endl;
}
//------------------------------------------------------------------------
DEFINE_ART_MODULE(novaddt::Merge2DTracks)
